"""
    :codeauthor: Thomas Jackson <jacksontj.89@gmail.com>
"""

import asyncio
import ctypes
import logging
import multiprocessing
import os
import threading
import time
import uuid

import pytest
import tornado.gen
import tornado.ioloop

import salt.channel.client
import salt.channel.server
import salt.crypt
import salt.transport.zeromq
import salt.utils.process
import salt.utils.stringutils
from salt.master import SMaster
from tests.conftest import FIPS_TESTRUN
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


pytestmark = [
    pytest.mark.core_test,
]


MASTER_PRIV_KEY = """
-----BEGIN RSA PRIVATE KEY-----
MIIEogIBAAKCAQEAoAsMPt+4kuIG6vKyw9r3+OuZrVBee/2vDdVetW+Js5dTlgrJ
aghWWn3doGmKlEjqh7E4UTa+t2Jd6w8RSLnyHNJ/HpVhMG0M07MF6FMfILtDrrt8
ZX7eDVt8sx5gCEpYI+XG8Y07Ga9i3Hiczt+fu6HYwu96HggmG2pqkOrn3iGfqBvV
YVFJzSZYe7e4c1PeEs0xYcrA4k+apyGsMtpef8vRUrNicRLc7dAcvfhtgt2DXEZ2
d72t/CR4ygtUvPXzisaTPW0G7OWAheCloqvTIIPQIjR8htFxGTz02STVXfnhnJ0Z
k8KhqKF2v1SQvIYxsZU7jaDgl5i3zpeh58cYOwIDAQABAoIBABZUJEO7Y91+UnfC
H6XKrZEZkcnH7j6/UIaOD9YhdyVKxhsnax1zh1S9vceNIgv5NltzIsfV6vrb6v2K
Dx/F7Z0O0zR5o+MlO8ZncjoNKskex10gBEWG00Uqz/WPlddiQ/TSMJTv3uCBAzp+
S2Zjdb4wYPUlgzSgb2ygxrhsRahMcSMG9PoX6klxMXFKMD1JxiY8QfAHahPzQXy9
F7COZ0fCVo6BE+MqNuQ8tZeIxu8mOULQCCkLFwXmkz1FpfK/kNRmhIyhxwvCS+z4
JuErW3uXfE64RLERiLp1bSxlDdpvRO2R41HAoNELTsKXJOEt4JANRHm/CeyA5wsh
NpscufUCgYEAxhgPfcMDy2v3nL6KtkgYjdcOyRvsAF50QRbEa8ldO+87IoMDD/Oe
osFERJ5hhyyEO78QnaLVegnykiw5DWEF02RKMhD/4XU+1UYVhY0wJjKQIBadsufB
2dnaKjvwzUhPh5BrBqNHl/FXwNCRDiYqXa79eWCPC9OFbZcUWWq70s8CgYEAztOI
61zRfmXJ7f70GgYbHg+GA7IrsAcsGRITsFR82Ho0lqdFFCxz7oK8QfL6bwMCGKyk
nzk+twh6hhj5UNp18KN8wktlo02zTgzgemHwaLa2cd6xKgmAyuPiTgcgnzt5LVNG
FOjIWkLwSlpkDTl7ZzY2QSy7t+mq5d750fpIrtUCgYBWXZUbcpPL88WgDB7z/Bjg
dlvW6JqLSqMK4b8/cyp4AARbNp12LfQC55o5BIhm48y/M70tzRmfvIiKnEc/gwaE
NJx4mZrGFFURrR2i/Xx5mt/lbZbRsmN89JM+iKWjCpzJ8PgIi9Wh9DIbOZOUhKVB
9RJEAgo70LvCnPTdS0CaVwKBgDJW3BllAvw/rBFIH4OB/vGnF5gosmdqp3oGo1Ik
jipmPAx6895AH4tquIVYrUl9svHsezjhxvjnkGK5C115foEuWXw0u60uiTiy+6Pt
2IS0C93VNMulenpnUrppE7CN2iWFAiaura0CY9fE/lsVpYpucHAWgi32Kok+ZxGL
WEttAoGAN9Ehsz4LeQxEj3x8wVeEMHF6OsznpwYsI2oVh6VxpS4AjgKYqeLVcnNi
TlZFsuQcqgod8OgzA91tdB+Rp86NygmWD5WzeKXpCOg9uA+y/YL+0sgZZHsuvbK6
PllUgXdYxqClk/hdBFB7v9AQoaj7K9Ga22v32msftYDQRJ94xOI=
-----END RSA PRIVATE KEY-----
"""


MASTER_PUB_KEY = """
-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAoAsMPt+4kuIG6vKyw9r3
+OuZrVBee/2vDdVetW+Js5dTlgrJaghWWn3doGmKlEjqh7E4UTa+t2Jd6w8RSLny
HNJ/HpVhMG0M07MF6FMfILtDrrt8ZX7eDVt8sx5gCEpYI+XG8Y07Ga9i3Hiczt+f
u6HYwu96HggmG2pqkOrn3iGfqBvVYVFJzSZYe7e4c1PeEs0xYcrA4k+apyGsMtpe
f8vRUrNicRLc7dAcvfhtgt2DXEZ2d72t/CR4ygtUvPXzisaTPW0G7OWAheCloqvT
IIPQIjR8htFxGTz02STVXfnhnJ0Zk8KhqKF2v1SQvIYxsZU7jaDgl5i3zpeh58cY
OwIDAQAB
-----END PUBLIC KEY-----
"""

MASTER2_PRIV_KEY = """
-----BEGIN RSA PRIVATE KEY-----
MIIEogIBAAKCAQEAp+8cTxguO6Vg+YO92VfHgNld3Zy8aM3JbZvpJcjTnis+YFJ7
Zlkcc647yPRRwY9nYBNywahnt5kIeuT1rTvTsMBZWvmUoEVUj1Xg8XXQkBvb9Ozy
Gqy/G/p8KDDpzMP/U+XCnUeHiXTZrgnqgBIc2cKeCVvWFqDi0GRFGzyaXLaX3PPm
M7DJ0MIPL1qgmcDq6+7Ze0gJ9SrDYFAeLmbuT1OqDfufXWQl/82JXeiwU2cOpqWq
7n5fvPOWim7l1tzQ+dSiMRRm0xa6uNexCJww3oJSwvMbAmgzvOhqqhlqv+K7u0u7
FrFFojESsL36Gq4GBrISnvu2tk7u4GGNTYYQbQIDAQABAoIBAADrqWDQnd5DVZEA
lR+WINiWuHJAy/KaIC7K4kAMBgbxrz2ZbiY9Ok/zBk5fcnxIZDVtXd1sZicmPlro
GuWodIxdPZAnWpZ3UtOXUayZK/vCP1YsH1agmEqXuKsCu6Fc+K8VzReOHxLUkmXn
FYM+tixGahXcjEOi/aNNTWitEB6OemRM1UeLJFzRcfyXiqzHpHCIZwBpTUAsmzcG
QiVDkMTKubwo/m+PVXburX2CGibUydctgbrYIc7EJvyx/cpRiPZXo1PhHQWdu4Y1
SOaC66WLsP/wqvtHo58JQ6EN/gjSsbAgGGVkZ1xMo66nR+pLpR27coS7o03xCks6
DY/0mukCgYEAuLIGgBnqoh7YsOBLd/Bc1UTfDMxJhNseo+hZemtkSXz2Jn51322F
Zw/FVN4ArXgluH+XsOhvG/MFFpojwZSrb0Qq5b1MRdo9qycq8lGqNtlN1WHqosDQ
zW29kpL0tlRrSDpww3wRESsN9rH5XIrJ1b3ZXuO7asR+KBVQMy/+NcUCgYEA6MSC
c+fywltKPgmPl5j0DPoDe5SXE/6JQy7w/vVGrGfWGf/zEJmhzS2R+CcfTTEqaT0T
Yw8+XbFgKAqsxwtE9MUXLTVLI3sSUyE4g7blCYscOqhZ8ItCUKDXWkSpt++rG0Um
1+cEJP/0oCazG6MWqvBC4NpQ1nzh46QpjWqMwokCgYAKDLXJ1p8rvx3vUeUJW6zR
dfPlEGCXuAyMwqHLxXgpf4EtSwhC5gSyPOtx2LqUtcrnpRmt6JfTH4ARYMW9TMef
QEhNQ+WYj213mKP/l235mg1gJPnNbUxvQR9lkFV8bk+AGJ32JRQQqRUTbU+yN2MQ
HEptnVqfTp3GtJIultfwOQKBgG+RyYmu8wBP650izg33BXu21raEeYne5oIqXN+I
R5DZ0JjzwtkBGroTDrVoYyuH1nFNEh7YLqeQHqvyufBKKYo9cid8NQDTu+vWr5UK
tGvHnwdKrJmM1oN5JOAiq0r7+QMAOWchVy449VNSWWV03aeftB685iR5BXkstbIQ
EVopAoGAfcGBTAhmceK/4Q83H/FXBWy0PAa1kZGg/q8+Z0KY76AqyxOVl0/CU/rB
3tO3sKhaMTHPME/MiQjQQGoaK1JgPY6JHYvly2KomrJ8QTugqNGyMzdVJkXAK2AM
GAwC8ivAkHf8CHrHa1W7l8t2IqBjW1aRt7mOW92nfG88Hck0Mbo=
-----END RSA PRIVATE KEY-----
"""


MASTER2_PUB_KEY = """
-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAp+8cTxguO6Vg+YO92VfH
gNld3Zy8aM3JbZvpJcjTnis+YFJ7Zlkcc647yPRRwY9nYBNywahnt5kIeuT1rTvT
sMBZWvmUoEVUj1Xg8XXQkBvb9OzyGqy/G/p8KDDpzMP/U+XCnUeHiXTZrgnqgBIc
2cKeCVvWFqDi0GRFGzyaXLaX3PPmM7DJ0MIPL1qgmcDq6+7Ze0gJ9SrDYFAeLmbu
T1OqDfufXWQl/82JXeiwU2cOpqWq7n5fvPOWim7l1tzQ+dSiMRRm0xa6uNexCJww
3oJSwvMbAmgzvOhqqhlqv+K7u0u7FrFFojESsL36Gq4GBrISnvu2tk7u4GGNTYYQ
bQIDAQAB
-----END PUBLIC KEY-----
"""


MASTER_SIGNING_PRIV = """
-----BEGIN RSA PRIVATE KEY-----
MIIEpAIBAAKCAQEAtieqrBMTM0MSIbhPKkDcozHqyXKyL/+bXYYw+iVPsns7c7bJ
zBqenLQlWoRVyrVyBFrrwQSrKu/0Mqn3l639iOGPlUoR3I7aZKIpyEdDkqd3xGIC
e+BtNNDqhUai67L63hEdG+iYAchi8UZw3LZGtcGpJ3FkBH4cYFX9EOam2QjbD7WY
EO7m1+j6XEYIOTCmAP9dGAvBbU0Jblc+wYxG3qNr+2dBWsK76QXWEqib2VSOGP+z
gjJa8tqY7PXXdOJpalQXNphmD/4o4pHKR4Euy0yL/1oMkpacmrV61LWB8Trnx9nS
9gdVrUteQF/cL1KAGwOsdVmiLpHfvqLLRqSAAQIDAQABAoIBABjB+HEN4Kixf4fk
wKHKEhL+SF6b/7sFX00NXZ/KLXRhSnnWSMQ8g/1hgMg2P2DfW4FbCDsCUu9xkLvI
HTZY+CJAIh9U42uaYPWXkt09TmJi76TZ+2Nx4/XvRUjbCm7Fs1I2ekHeUbbAUS5g
+BsPjTnL+h05zLHNoDa5yT0gVGIgFsQcX/w38arZCe8Rjp9le7PXUB5IIqASsDiw
t8zJvdyWToeXd0WswCHTQu5coHvKo5MCjIZZ1Ink1yJcCCc3rKDc+q3jB2z9T9oW
cUsKzJ4VuleiYj1eRxFITBmXbjKrb/GPRRUkeqCQbs68Hyj2d3UtOFDPeF4vng/3
jGsHPq8CgYEA0AHAbwykVC6NMa37BTvEqcKoxbjTtErxR+yczlmVDfma9vkwtZvx
FJdbS/+WGA/ucDby5x5b2T5k1J9ueMR86xukb+HnyS0WKsZ94Ie8WnJAcbp+38M6
7LD0u74Cgk93oagDAzUHqdLq9cXxv/ppBpxVB1Uvu8DfVMHj+wt6ie8CgYEA4C7u
u+6b8EmbGqEdtlPpScKG0WFstJEDGXRARDCRiVP2w6wm25v8UssCPvWcwf8U1Hoq
lhMY+H6a5dnRRiNYql1MGQAsqMi7VeJNYb0B1uxi7X8MPM+SvXoAglX7wm1z0cVy
O4CE5sEKbBg6aQabx1x9tzdrm80SKuSsLc5HRQ8CgYEAp/mCKSuQWNru8ruJBwTp
IB4upN1JOUN77ZVKW+lD0XFMjz1U9JPl77b65ziTQQM8jioRpkqB6cHVM088qxIh
vssn06Iex/s893YrmPKETJYPLMhqRNEn+JQ+To53ADykY0uGg0SD18SYMbmULHBP
+CKvF6jXT0vGDnA1ZzoxzskCgYEA2nQhYrRS9EVlhP93KpJ+A8gxA5tCCHo+YPFt
JoWFbCKLlYUNoHZR3IPCPoOsK0Zbj+kz0mXtsUf9vPkR+py669haLQqEejyQgFIz
QYiiYEKc6/0feapzvXtDP751w7JQaBtVAzJrT0jQ1SCO2oT8C7rPLlgs3fdpOq72
MPSPcnUCgYBWHm6bn4HvaoUSr0v2hyD9fHZS/wDTnlXVe5c1XXgyKlJemo5dvycf
HUCmN/xIuO6AsiMdqIzv+arNJdboz+O+bNtS43LkTJfEH3xj2/DdUogdvOgG/iPM
u9KBT1h+euws7PqC5qt4vqLwCTTCZXmUS8Riv+62RCC3kZ5AbpT3ZA==
-----END RSA PRIVATE KEY-----
"""

MASTER_SIGNING_PUB = """
-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAtieqrBMTM0MSIbhPKkDc
ozHqyXKyL/+bXYYw+iVPsns7c7bJzBqenLQlWoRVyrVyBFrrwQSrKu/0Mqn3l639
iOGPlUoR3I7aZKIpyEdDkqd3xGICe+BtNNDqhUai67L63hEdG+iYAchi8UZw3LZG
tcGpJ3FkBH4cYFX9EOam2QjbD7WYEO7m1+j6XEYIOTCmAP9dGAvBbU0Jblc+wYxG
3qNr+2dBWsK76QXWEqib2VSOGP+zgjJa8tqY7PXXdOJpalQXNphmD/4o4pHKR4Eu
y0yL/1oMkpacmrV61LWB8Trnx9nS9gdVrUteQF/cL1KAGwOsdVmiLpHfvqLLRqSA
AQIDAQAB
-----END PUBLIC KEY-----
"""

MINION_PRIV_KEY = """
-----BEGIN RSA PRIVATE KEY-----
MIIEowIBAAKCAQEAsT6TwnlI0L7urjXu6D5E11tFJ/NglQ45jW/WN9tAUNvphq6Q
cjJCd/aWmdqlqe7ix8y9M/8rgwghRQsnPXblVBvPwFcUEXhMRnOGzqbq/0zyQX01
KecT0plBhlDt2lTyCLU6E4XCqyLbPfOxgXzsVqM0/TnzRtpVvGNy+5N4eFGylrjb
cJhPxKt2G9TDOCM/hYacDs5RVIYQQmcYb8LJq7G3++FfWpYRDaxdKoHNFDspEynd
jzr67hgThnwzc388OKNJx/7B2atwPTunPb3YBjgwDyRO/01OKK4gUHdw5KoctFgp
kDCDjwjemlyXV+MYODRTIdtOlAP83ZkntEuLoQIDAQABAoIBAAJOKNtvFGfF2l9H
S4CXZSUGU0a+JaCkR+wmnjsPwPn/dXDpAe8nGpidpNicPWqRm6WABjeQHaxda+fB
lpSrRtEdo3zoi2957xQJ5wddDtI1pmXJQrdbm0H/K39oIg/Xtv/IZT769TM6OtVg
paUxG/aftmeGXDtGfIL8w1jkuPABRBLOakWQA9uVdeG19KTU0Ag8ilpJdEX64uFJ
W75bpVjT+KO/6aV1inuCntQSP097aYvUWajRwuiYVJOxoBZHme3IObcE6mdnYXeQ
wblyWBpJUHrOS4MP4HCODV2pHKZ2rr7Nwhh8lMNw/eY9OP0ifz2AcAqe3sUMQOKP
T0qRC6ECgYEAyeU5JvUPOpxXvvChYh6gJ8pYTIh1ueDP0O5e4t3vhz6lfy9DKtRN
ROJLUorHvw/yVXMR72nT07a0z2VswcrUSw8ov3sI53F0NkLGEafQ35lVhTGs4vTl
CFoQCuAKPsxeUl4AIbfbpkDsLGQqzW1diFArK7YeQkpGuGaGodXl480CgYEA4L40
x5cUXnAhTPsybo7sbcpiwFHoGblmdkvpYvHA2QxtNSi2iHHdqGo8qP1YsZjKQn58
371NhtqidrJ6i/8EBFP1dy+y/jr9qYlZNNGcQeBi+lshrEOIf1ct56KePG79s8lm
DmD1OY8tO2R37+Py46Nq1n6viT/ST4NjLQI3GyUCgYEAiOswSDA3ZLs0cqRD/gPg
/zsliLmehTFmHj4aEWcLkz+0Ar3tojUaNdX12QOPFQ7efH6uMhwl8NVeZ6xUBlTk
hgbAzqLE1hjGBCpiowSZDZqyOcMHiV8ll/VkHcv0hsQYT2m6UyOaDXTH9g70TB6Y
KOKddGZsvO4cad/1+/jQkB0CgYAzDEEkzLY9tS57M9uCrUgasAu6L2CO50PUvu1m
Ig9xvZbYqkS7vVFhva/FmrYYsOHQNLbcgz0m0mZwm52mSuh4qzFoPxdjE7cmWSJA
ExRxCiyxPR3q6PQKKJ0urgtPIs7RlX9u6KsKxfC6OtnbTWWQO0A7NE9e13ZHxUoz
oPsvWQKBgCa0+Fb2lzUeiQz9bV1CBkWneDZUXuZHmabAZomokX+h/bq+GcJFzZjW
3kAHwYkIy9IAy3SyO/6CP0V3vAye1p+XbotiwsQ/XZnr0pflSQL3J1l1CyN3aopg
Niv7k/zBn15B72aK73R/CpUSk9W/eJGqk1NcNwf8hJHsboRYx6BR
-----END RSA PRIVATE KEY-----
"""


MINION_PUB_KEY = """
-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAsT6TwnlI0L7urjXu6D5E
11tFJ/NglQ45jW/WN9tAUNvphq6QcjJCd/aWmdqlqe7ix8y9M/8rgwghRQsnPXbl
VBvPwFcUEXhMRnOGzqbq/0zyQX01KecT0plBhlDt2lTyCLU6E4XCqyLbPfOxgXzs
VqM0/TnzRtpVvGNy+5N4eFGylrjbcJhPxKt2G9TDOCM/hYacDs5RVIYQQmcYb8LJ
q7G3++FfWpYRDaxdKoHNFDspEyndjzr67hgThnwzc388OKNJx/7B2atwPTunPb3Y
BjgwDyRO/01OKK4gUHdw5KoctFgpkDCDjwjemlyXV+MYODRTIdtOlAP83ZkntEuL
oQIDAQAB
-----END PUBLIC KEY-----
"""

AES_KEY = "8wxWlOaMMQ4d3yT74LL4+hGrGTf65w8VgrcNjLJeLRQ2Q6zMa8ItY2EQUgMKKDb7JY+RnPUxbB0="


@pytest.fixture
def signing_algorithm():
    if FIPS_TESTRUN:
        return salt.crypt.PKCS1v15_SHA224
    return salt.crypt.PKCS1v15_SHA1


@pytest.fixture
def encryption_algorithm():
    if FIPS_TESTRUN:
        return salt.crypt.OAEP_SHA224
    return salt.crypt.OAEP_SHA1


@pytest.fixture
def pki_dir(tmp_path):
    _pki_dir = tmp_path / "pki"
    _pki_dir.mkdir()
    madir = _pki_dir / "master"
    madir.mkdir()

    mapriv = madir / "master.pem"
    mapriv.write_text(MASTER_PRIV_KEY.strip())
    mapub = madir / "master.pub"
    mapub.write_text(MASTER_PUB_KEY.strip())

    maspriv = madir / "master_sign.pem"
    maspriv.write_text(MASTER_SIGNING_PRIV.strip())
    maspub = madir / "master_sign.pub"
    maspub.write_text(MASTER_SIGNING_PUB.strip())

    misdir = madir / "minions"
    misdir.mkdir()
    misdir.joinpath("minion").write_text(MINION_PUB_KEY.strip())
    for sdir in [
        "minions_autosign",
        "minions_denied",
        "minions_pre",
        "minions_rejected",
    ]:
        madir.joinpath(sdir).mkdir()

    midir = _pki_dir / "minion"
    midir.mkdir()
    mipub = midir / "minion.pub"
    mipub.write_text(MINION_PUB_KEY.strip())
    mipriv = midir / "minion.pem"
    mipriv.write_text(MINION_PRIV_KEY.strip())
    mimapriv = midir / "minion_master.pub"
    mimapriv.write_text(MASTER_PUB_KEY.strip())
    mimaspriv = midir / "master_sign.pub"
    mimaspriv.write_text(MASTER_SIGNING_PUB.strip())
    yield _pki_dir


def test_master_uri():
    """
    test _get_master_uri method
    """

    m_ip = "127.0.0.1"
    m_port = 4505
    s_ip = "111.1.0.1"
    s_port = 4058

    m_ip6 = "1234:5678::9abc"
    s_ip6 = "1234:5678::1:9abc"

    with patch("salt.transport.zeromq.LIBZMQ_VERSION_INFO", (4, 1, 6)), patch(
        "salt.transport.zeromq.ZMQ_VERSION_INFO", (16, 0, 1)
    ):
        # pass in both source_ip and source_port
        assert (
            salt.transport.zeromq._get_master_uri(
                master_ip=m_ip, master_port=m_port, source_ip=s_ip, source_port=s_port
            )
            == f"tcp://{s_ip}:{s_port};{m_ip}:{m_port}"
        )

        assert (
            salt.transport.zeromq._get_master_uri(
                master_ip=m_ip6, master_port=m_port, source_ip=s_ip6, source_port=s_port
            )
            == f"tcp://[{s_ip6}]:{s_port};[{m_ip6}]:{m_port}"
        )

        # source ip and source_port empty
        assert (
            salt.transport.zeromq._get_master_uri(master_ip=m_ip, master_port=m_port)
            == f"tcp://{m_ip}:{m_port}"
        )

        assert (
            salt.transport.zeromq._get_master_uri(master_ip=m_ip6, master_port=m_port)
            == f"tcp://[{m_ip6}]:{m_port}"
        )

        # pass in only source_ip
        assert (
            salt.transport.zeromq._get_master_uri(
                master_ip=m_ip, master_port=m_port, source_ip=s_ip
            )
            == f"tcp://{s_ip}:0;{m_ip}:{m_port}"
        )

        assert (
            salt.transport.zeromq._get_master_uri(
                master_ip=m_ip6, master_port=m_port, source_ip=s_ip6
            )
            == f"tcp://[{s_ip6}]:0;[{m_ip6}]:{m_port}"
        )

        # pass in only source_port
        assert (
            salt.transport.zeromq._get_master_uri(
                master_ip=m_ip, master_port=m_port, source_port=s_port
            )
            == f"tcp://0.0.0.0:{s_port};{m_ip}:{m_port}"
        )


def test_clear_req_channel_master_uri_override(temp_salt_minion, temp_salt_master):
    """
    ensure master_uri kwarg is respected
    """
    opts = temp_salt_minion.config.copy()
    # minion_config should be 127.0.0.1, we want a different uri that still connects
    opts.update(
        {
            "id": "root",
            "transport": "zeromq",
            "auth_tries": 1,
            "auth_timeout": 5,
            "master_ip": "127.0.0.1",
            "master_port": temp_salt_master.config["ret_port"],
            "master_uri": "tcp://127.0.0.1:{}".format(
                temp_salt_master.config["ret_port"]
            ),
        }
    )
    master_uri = "tcp://{master_ip}:{master_port}".format(
        master_ip="localhost", master_port=opts["master_port"]
    )
    with salt.channel.client.ReqChannel.factory(opts, master_uri=master_uri) as channel:
        assert "127.0.0.1" in channel.transport.master_uri


def run_loop_in_thread(loop, evt):
    """
    Run the provided loop until an event is set
    """
    asyncio.set_event_loop(loop.asyncio_loop)

    async def stopper():
        await asyncio.sleep(0.1)
        while True:
            if not evt.is_set():
                loop.stop()
                break
            await asyncio.sleep(0.3)

    loop.add_callback(evt.set)
    loop.add_callback(stopper)
    try:
        loop.start()
    finally:
        loop.close()


class MockSaltMinionMaster:
    mock = MagicMock()

    def __init__(self, temp_salt_minion, temp_salt_master):
        SMaster.secrets["aes"] = {
            "secret": multiprocessing.Array(
                ctypes.c_char,
                salt.utils.stringutils.to_bytes(
                    salt.crypt.Crypticle.generate_key_string()
                ),
            ),
            "reload": salt.crypt.Crypticle.generate_key_string,
        }
        self.process_manager = salt.utils.process.ProcessManager(
            name="ReqServer_ProcessManager"
        )

        master_opts = temp_salt_master.config.copy()
        master_opts.update({"transport": "zeromq"})
        self.server_channel = salt.channel.server.ReqServerChannel.factory(master_opts)
        self.server_channel.pre_fork(self.process_manager)
        self.io_loop = tornado.ioloop.IOLoop(make_current=False)
        self.evt = threading.Event()
        self.server_channel.post_fork(self._handle_payload, io_loop=self.io_loop)
        self.server_thread = threading.Thread(
            target=run_loop_in_thread, args=(self.io_loop, self.evt)
        )
        self.server_thread.start()
        minion_opts = temp_salt_minion.config.copy()
        minion_opts.update(
            {
                "master_ip": "127.0.0.1",
                "transport": "zeromq",
            }
        )
        self.channel = salt.channel.client.ReqChannel.factory(
            minion_opts, crypt="clear"
        )

    def __enter__(self):
        self.channel.__enter__()
        self.evt.wait()
        return self

    def __exit__(self, *args, **kwargs):
        self.channel.__exit__(*args, **kwargs)
        del self.channel
        self.server_channel.close()
        # Attempting to kill the children hangs the test suite.
        # Let the test suite handle this instead.
        self.process_manager.stop_restarting()
        self.process_manager.kill_children()
        self.evt.clear()
        self.server_thread.join()
        # Give the procs a chance to fully close before we stop the io_loop
        time.sleep(2)
        SMaster.secrets.pop("aes")
        del self.server_channel
        del self.io_loop
        del self.process_manager
        del self.server_thread

    # pylint: enable=W1701
    @classmethod
    @tornado.gen.coroutine
    def _handle_payload(cls, payload):
        """
        TODO: something besides echo
        """
        cls.mock._handle_payload_hook()
        raise tornado.gen.Return((payload, {"fun": "send_clear"}))


@pytest.mark.parametrize("message", ["", [], ()])
def test_badload(temp_salt_minion, temp_salt_master, message):
    """
    Test a variety of bad requests, make sure that we get some sort of error
    """
    with MockSaltMinionMaster(temp_salt_minion, temp_salt_master) as minion_master:
        ret = minion_master.channel.send(message, timeout=5, tries=1)
        assert ret == "payload and load must be a dict"


def test_payload_handling_exception(temp_salt_minion, temp_salt_master):
    """
    test of getting exception on payload handling
    """
    with MockSaltMinionMaster(temp_salt_minion, temp_salt_master) as minion_master:
        with patch.object(minion_master.mock, "_handle_payload_hook") as _mock:
            _mock.side_effect = Exception()
            ret = minion_master.channel.send({}, timeout=15, tries=1)
            assert ret == "Some exception handling minion payload"


def test_serverside_exception(temp_salt_minion, temp_salt_master):
    """
    test of getting server side exception on payload handling
    """
    with MockSaltMinionMaster(temp_salt_minion, temp_salt_master) as minion_master:
        with patch.object(minion_master.mock, "_handle_payload_hook") as _mock:
            _mock.side_effect = tornado.gen.Return(({}, {"fun": "madeup-fun"}))
            ret = minion_master.channel.send({}, timeout=5, tries=1)
            assert ret == "Server-side exception handling payload"


def test_req_server_chan_encrypt_v2(
    pki_dir, encryption_algorithm, signing_algorithm, master_opts
):
    loop = tornado.ioloop.IOLoop.current()
    master_opts.update(
        {
            "worker_threads": 1,
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "zmq_monitor": False,
            "mworker_queue_niceness": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("master")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
        }
    )
    server = salt.channel.server.ReqServerChannel.factory(master_opts)
    dictkey = "pillar"
    nonce = "abcdefg"
    pillar_data = {"pillar1": "meh"}
    try:
        ret = server._encrypt_private(
            pillar_data,
            dictkey,
            "minion",
            nonce,
            encryption_algorithm=encryption_algorithm,
            signing_algorithm=signing_algorithm,
        )
        assert "key" in ret
        assert dictkey in ret

        key = salt.crypt.PrivateKey(str(pki_dir.joinpath("minion", "minion.pem")))
        aes = key.decrypt(ret["key"], encryption_algorithm)
        pcrypt = salt.crypt.Crypticle(master_opts, aes)
        signed_msg = pcrypt.loads(ret[dictkey])

        assert "sig" in signed_msg
        assert "data" in signed_msg
        data = salt.payload.loads(signed_msg["data"])
        assert "key" in data
        assert data["key"] == ret["key"]
        assert "key" in data
        assert data["nonce"] == nonce
        assert "pillar" in data
        assert data["pillar"] == pillar_data
    finally:
        server.close()


def test_req_server_chan_encrypt_v1(pki_dir, encryption_algorithm, master_opts):
    loop = tornado.ioloop.IOLoop.current()
    master_opts.update(
        {
            "worker_threads": 1,
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "zmq_monitor": False,
            "mworker_queue_niceness": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("master")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
        }
    )
    server = salt.channel.server.ReqServerChannel.factory(master_opts)
    dictkey = "pillar"
    nonce = "abcdefg"
    pillar_data = {"pillar1": "meh"}
    try:
        ret = server._encrypt_private(
            pillar_data,
            dictkey,
            "minion",
            sign_messages=False,
            encryption_algorithm=encryption_algorithm,
        )

        assert "key" in ret
        assert dictkey in ret

        key = salt.crypt.PrivateKey(str(pki_dir.joinpath("minion", "minion.pem")))
        aes = key.decrypt(ret["key"], encryption_algorithm)
        pcrypt = salt.crypt.Crypticle(master_opts, aes)
        data = pcrypt.loads(ret[dictkey])
        assert data == pillar_data
    finally:
        server.close()


def test_req_chan_decode_data_dict_entry_v1(
    pki_dir, encryption_algorithm, minion_opts, master_opts
):
    mockloop = MagicMock()
    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "acceptance_wait_time": 3,
            "acceptance_wait_time_max": 3,
        }
    )
    master_opts = dict(master_opts, pki_dir=str(pki_dir.joinpath("master")))
    server = salt.channel.server.ReqServerChannel.factory(master_opts)
    client = salt.channel.client.ReqChannel.factory(minion_opts, io_loop=mockloop)
    try:
        dictkey = "pillar"
        target = "minion"
        pillar_data = {"pillar1": "meh"}
        ret = server._encrypt_private(
            pillar_data,
            dictkey,
            target,
            sign_messages=False,
            encryption_algorithm=encryption_algorithm,
        )
        key = client.auth.get_keys()
        aes = key.decrypt(ret["key"], encryption_algorithm)
        pcrypt = salt.crypt.Crypticle(client.opts, aes)
        ret_pillar_data = pcrypt.loads(ret[dictkey])
        assert ret_pillar_data == pillar_data
    finally:
        client.close()
        server.close()


async def test_req_chan_decode_data_dict_entry_v2(minion_opts, master_opts, pki_dir):
    mockloop = MagicMock()
    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "acceptance_wait_time": 3,
            "acceptance_wait_time_max": 3,
        }
    )
    master_opts.update(pki_dir=str(pki_dir.joinpath("master")))
    server = salt.channel.server.ReqServerChannel.factory(master_opts)
    client = salt.channel.client.AsyncReqChannel.factory(minion_opts, io_loop=mockloop)

    dictkey = "pillar"
    target = "minion"
    pillar_data = {"pillar1": "meh"}

    # Mock auth and message client.
    auth = client.auth
    auth._crypticle = salt.crypt.Crypticle(minion_opts, AES_KEY)
    client.auth = MagicMock()
    client.auth.mpub = auth.mpub
    client.auth.authenticated = True
    client.auth.get_keys = auth.get_keys
    client.auth.crypticle.dumps = auth.crypticle.dumps
    client.auth.crypticle.loads = auth.crypticle.loads
    client.transport = MagicMock()

    print(minion_opts["encryption_algorithm"])

    @tornado.gen.coroutine
    def mocksend(msg, timeout=60, tries=3):
        client.transport.msg = msg
        load = client.auth.crypticle.loads(msg["load"])
        ret = server._encrypt_private(
            pillar_data,
            dictkey,
            target,
            nonce=load["nonce"],
            sign_messages=True,
            encryption_algorithm=minion_opts["encryption_algorithm"],
            signing_algorithm=minion_opts["signing_algorithm"],
        )
        raise tornado.gen.Return(ret)

    client.transport.send = mocksend

    # Note the 'ver' value in 'load' does not represent the the 'version' sent
    # in the top level of the transport's message.
    load = {
        "id": target,
        "grains": {},
        "saltenv": "base",
        "pillarenv": "base",
        "pillar_override": True,
        "extra_minion_data": {},
        "ver": "2",
        "cmd": "_pillar",
    }
    try:
        ret = await client.crypted_transfer_decode_dictentry(  # pylint: disable=E1121,E1123
            load,
            dictkey="pillar",
        )
        assert "version" in client.transport.msg
        assert client.transport.msg["version"] == 2
        assert ret == {"pillar1": "meh"}
    finally:
        client.close()
        server.close()


async def test_req_chan_decode_data_dict_entry_v2_bad_nonce(
    minion_opts, master_opts, pki_dir
):
    mockloop = MagicMock()
    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "acceptance_wait_time": 3,
            "acceptance_wait_time_max": 3,
        }
    )
    master_opts.update(pki_dir=str(pki_dir.joinpath("master")))
    server = salt.channel.server.ReqServerChannel.factory(master_opts)
    client = salt.channel.client.AsyncReqChannel.factory(minion_opts, io_loop=mockloop)

    dictkey = "pillar"
    badnonce = "abcdefg"
    target = "minion"
    pillar_data = {"pillar1": "meh"}

    # Mock auth and message client.
    auth = client.auth
    auth._crypticle = salt.crypt.Crypticle(minion_opts, AES_KEY)
    client.auth = MagicMock()
    client.auth.mpub = auth.mpub
    client.auth.authenticated = True
    client.auth.get_keys = auth.get_keys
    client.auth.crypticle.dumps = auth.crypticle.dumps
    client.auth.crypticle.loads = auth.crypticle.loads
    real_transport = client.transport
    client.transport = MagicMock()
    real_transport.close()
    ret = server._encrypt_private(
        pillar_data,
        dictkey,
        target,
        nonce=badnonce,
        sign_messages=True,
        encryption_algorithm=minion_opts["encryption_algorithm"],
        signing_algorithm=minion_opts["signing_algorithm"],
    )

    @tornado.gen.coroutine
    def mocksend(msg, timeout=60, tries=3):
        client.transport.msg = msg
        raise tornado.gen.Return(ret)

    client.transport.send = mocksend

    # Note the 'ver' value in 'load' does not represent the the 'version' sent
    # in the top level of the transport's message.
    load = {
        "id": target,
        "grains": {},
        "saltenv": "base",
        "pillarenv": "base",
        "pillar_override": True,
        "extra_minion_data": {},
        "ver": "2",
        "cmd": "_pillar",
    }

    try:
        with pytest.raises(salt.crypt.AuthenticationError) as excinfo:
            ret = await client.crypted_transfer_decode_dictentry(  # pylint: disable=E1121,E1123
                load,
                dictkey="pillar",
            )
        assert "Pillar nonce verification failed." == excinfo.value.message
    finally:
        client.close()
        server.close()


async def test_req_chan_decode_data_dict_entry_v2_bad_signature(
    pki_dir, minion_opts, master_opts
):
    mockloop = MagicMock()
    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "acceptance_wait_time": 3,
            "acceptance_wait_time_max": 3,
        }
    )
    master_opts.update(pki_dir=str(pki_dir.joinpath("master")))
    server = salt.channel.server.ReqServerChannel.factory(master_opts)
    client = salt.channel.client.AsyncReqChannel.factory(minion_opts, io_loop=mockloop)

    dictkey = "pillar"
    badnonce = "abcdefg"
    target = "minion"
    pillar_data = {"pillar1": "meh"}

    # Mock auth and message client.
    auth = client.auth
    auth._crypticle = salt.crypt.Crypticle(minion_opts, AES_KEY)
    client.auth = MagicMock()
    client.auth.mpub = auth.mpub
    client.auth.authenticated = True
    client.auth.get_keys = auth.get_keys
    client.auth.crypticle.dumps = auth.crypticle.dumps
    client.auth.crypticle.loads = auth.crypticle.loads
    client.transport = MagicMock()

    @tornado.gen.coroutine
    def mocksend(msg, timeout=60, tries=3):
        client.transport.msg = msg
        load = client.auth.crypticle.loads(msg["load"])
        ret = server._encrypt_private(
            pillar_data,
            dictkey,
            target,
            nonce=load["nonce"],
            sign_messages=True,
            encryption_algorithm=minion_opts["encryption_algorithm"],
            signing_algorithm=minion_opts["signing_algorithm"],
        )

        key = client.auth.get_keys()
        aes = key.decrypt(ret["key"], minion_opts["encryption_algorithm"])
        pcrypt = salt.crypt.Crypticle(client.opts, aes)
        signed_msg = pcrypt.loads(ret[dictkey])
        # Changing the pillar data will cause the signature verification to
        # fail.
        data = salt.payload.loads(signed_msg["data"])
        data["pillar"] = {"pillar1": "bar"}
        signed_msg["data"] = salt.payload.dumps(data)
        ret[dictkey] = pcrypt.dumps(signed_msg)
        raise tornado.gen.Return(ret)

    client.transport.send = mocksend

    # Note the 'ver' value in 'load' does not represent the the 'version' sent
    # in the top level of the transport's message.
    load = {
        "id": target,
        "grains": {},
        "saltenv": "base",
        "pillarenv": "base",
        "pillar_override": True,
        "extra_minion_data": {},
        "ver": "2",
        "cmd": "_pillar",
    }

    try:
        with pytest.raises(salt.crypt.AuthenticationError) as excinfo:
            ret = await client.crypted_transfer_decode_dictentry(  # pylint: disable=E1121,E1123
                load,
                dictkey="pillar",
            )
        assert "Pillar payload signature failed to validate." == excinfo.value.message
    finally:
        client.close()
        server.close()


async def test_req_chan_decode_data_dict_entry_v2_bad_key(
    pki_dir, minion_opts, master_opts
):
    mockloop = MagicMock()
    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "acceptance_wait_time": 3,
            "acceptance_wait_time_max": 3,
        }
    )
    master_opts.update(pki_dir=str(pki_dir.joinpath("master")))
    server = salt.channel.server.ReqServerChannel.factory(master_opts)
    client = salt.channel.client.AsyncReqChannel.factory(minion_opts, io_loop=mockloop)

    dictkey = "pillar"
    badnonce = "abcdefg"
    target = "minion"
    pillar_data = {"pillar1": "meh"}

    # Mock auth and message client.
    auth = client.auth
    auth._crypticle = salt.crypt.Crypticle(master_opts, AES_KEY)
    client.auth = MagicMock()
    client.auth.mpub = auth.mpub
    client.auth.authenticated = True
    client.auth.get_keys = auth.get_keys
    client.auth.crypticle.dumps = auth.crypticle.dumps
    client.auth.crypticle.loads = auth.crypticle.loads
    client.transport = MagicMock()

    @tornado.gen.coroutine
    def mocksend(msg, timeout=60, tries=3):
        client.transport.msg = msg
        load = client.auth.crypticle.loads(msg["load"])
        ret = server._encrypt_private(
            pillar_data,
            dictkey,
            target,
            nonce=load["nonce"],
            sign_messages=True,
            encryption_algorithm=minion_opts["encryption_algorithm"],
            signing_algorithm=minion_opts["signing_algorithm"],
        )

        mkey = client.auth.get_keys()
        aes = mkey.decrypt(ret["key"], minion_opts["encryption_algorithm"])
        pcrypt = salt.crypt.Crypticle(client.opts, aes)
        signed_msg = pcrypt.loads(ret[dictkey])

        # Now encrypt with a different key
        key = salt.crypt.Crypticle.generate_key_string()
        pcrypt = salt.crypt.Crypticle(master_opts, key)
        pubfn = os.path.join(master_opts["pki_dir"], "minions", "minion")
        pub = salt.crypt.PublicKey(pubfn)
        ret[dictkey] = pcrypt.dumps(signed_msg)
        key = salt.utils.stringutils.to_bytes(key)
        ret["key"] = pub.encrypt(key, minion_opts["encryption_algorithm"])
        raise tornado.gen.Return(ret)

    client.transport.send = mocksend

    # Note the 'ver' value in 'load' does not represent the the 'version' sent
    # in the top level of the transport's message.
    load = {
        "id": target,
        "grains": {},
        "saltenv": "base",
        "pillarenv": "base",
        "pillar_override": True,
        "extra_minion_data": {},
        "ver": "2",
        "cmd": "_pillar",
    }
    try:
        with pytest.raises(salt.crypt.AuthenticationError) as excinfo:
            await client.crypted_transfer_decode_dictentry(  # pylint: disable=E1121,E1123
                load,
                dictkey="pillar",
            )
        assert "Key verification failed." == excinfo.value.message
    finally:
        client.close()
        server.close()


async def test_req_serv_auth_v1(minion_opts, master_opts, pki_dir):
    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "max_minions": 0,
            "auto_accept": False,
            "open_mode": False,
            "key_pass": None,
            "master_sign_pubkey": False,
            "publish_port": 4505,
            "auth_mode": 1,
        }
    )
    SMaster.secrets["aes"] = {
        "secret": multiprocessing.Array(
            ctypes.c_char,
            salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
        ),
        "reload": salt.crypt.Crypticle.generate_key_string,
    }
    master_opts.update(pki_dir=str(pki_dir.joinpath("master")))
    server = salt.channel.server.ReqServerChannel.factory(master_opts)
    server.auto_key = salt.daemons.masterapi.AutoKey(server.opts)
    server.cache_cli = False
    server.master_key = salt.crypt.MasterKeys(server.opts)

    pub = salt.crypt.get_rsa_pub_key(str(pki_dir.joinpath("minion", "minion.pub")))
    token = salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string())
    nonce = uuid.uuid4().hex

    # We need to read the public key with fopen otherwise the newlines might
    # not match on windows.
    with salt.utils.files.fopen(
        str(pki_dir.joinpath("minion", "minion.pub")), "r"
    ) as fp:
        pub_key = salt.crypt.clean_key(fp.read())

    load = {
        "cmd": "_auth",
        "id": "minion",
        "token": token,
        "pub": pub_key,
        "enc_algo": minion_opts["encryption_algorithm"],
        "sig_algo": minion_opts["signing_algorithm"],
    }
    try:
        ret = server._auth(load, sign_messages=False)
        assert "load" not in ret
    finally:
        server.close()


async def test_req_serv_auth_v2(minion_opts, master_opts, pki_dir):
    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "max_minions": 0,
            "auto_accept": False,
            "open_mode": False,
            "key_pass": None,
            "master_sign_pubkey": False,
            "publish_port": 4505,
            "auth_mode": 1,
        }
    )
    SMaster.secrets["aes"] = {
        "secret": multiprocessing.Array(
            ctypes.c_char,
            salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
        ),
        "reload": salt.crypt.Crypticle.generate_key_string,
    }
    master_opts.update(pki_dir=str(pki_dir.joinpath("master")))
    server = salt.channel.server.ReqServerChannel.factory(master_opts)
    server.auto_key = salt.daemons.masterapi.AutoKey(server.opts)
    server.cache_cli = False
    server.master_key = salt.crypt.MasterKeys(server.opts)

    pub = salt.crypt.get_rsa_pub_key(str(pki_dir.joinpath("minion", "minion.pub")))
    token = salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string())
    nonce = uuid.uuid4().hex

    # We need to read the public key with fopen otherwise the newlines might
    # not match on windows.
    with salt.utils.files.fopen(
        str(pki_dir.joinpath("minion", "minion.pub")), "r"
    ) as fp:
        pub_key = salt.crypt.clean_key(fp.read())

    load = {
        "cmd": "_auth",
        "id": "minion",
        "nonce": nonce,
        "token": token,
        "pub": pub_key,
        "enc_algo": minion_opts["encryption_algorithm"],
        "sig_algo": minion_opts["signing_algorithm"],
    }
    try:
        ret = server._auth(load, sign_messages=True)
        assert "sig" in ret
        assert "load" in ret
    finally:
        server.close()


async def test_req_chan_auth_v2(minion_opts, master_opts, pki_dir, io_loop):
    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "max_minions": 0,
            "auto_accept": False,
            "open_mode": False,
            "key_pass": None,
            "publish_port": 4505,
            "auth_mode": 1,
            "acceptance_wait_time": 3,
            "acceptance_wait_time_max": 3,
        }
    )
    SMaster.secrets["aes"] = {
        "secret": multiprocessing.Array(
            ctypes.c_char,
            salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
        ),
        "reload": salt.crypt.Crypticle.generate_key_string,
    }
    master_opts.update(pki_dir=str(pki_dir.joinpath("master")))
    master_opts["master_sign_pubkey"] = False
    server = salt.channel.server.ReqServerChannel.factory(master_opts)
    server.auto_key = salt.daemons.masterapi.AutoKey(server.opts)
    server.cache_cli = False
    server.master_key = salt.crypt.MasterKeys(server.opts)
    minion_opts["verify_master_pubkey_sign"] = False
    minion_opts["always_verify_signature"] = False
    client = salt.channel.client.AsyncReqChannel.factory(minion_opts, io_loop=io_loop)
    signin_payload = client.auth.minion_sign_in_payload()
    pload = client._package_load(signin_payload)
    try:
        assert "version" in pload
        assert pload["version"] == 2

        ret = server._auth(pload["load"], sign_messages=True)
        assert "sig" in ret
        ret = client.auth.handle_signin_response(signin_payload, ret)
        assert "aes" in ret
        assert "master_uri" in ret
        assert "publish_port" in ret
    finally:
        client.close()
        server.close()


async def test_req_chan_auth_v2_with_master_signing(
    pki_dir, io_loop, minion_opts, master_opts
):
    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "max_minions": 0,
            "auto_accept": False,
            "open_mode": False,
            "key_pass": None,
            "publish_port": 4505,
            "auth_mode": 1,
            "acceptance_wait_time": 3,
            "acceptance_wait_time_max": 3,
        }
    )
    SMaster.secrets["aes"] = {
        "secret": multiprocessing.Array(
            ctypes.c_char,
            salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
        ),
        "reload": salt.crypt.Crypticle.generate_key_string,
    }
    master_opts = dict(master_opts, pki_dir=str(pki_dir.joinpath("master")))
    master_opts["master_sign_pubkey"] = True
    master_opts["master_use_pubkey_signature"] = False
    master_opts["signing_key_pass"] = ""
    master_opts["master_sign_key_name"] = "master_sign"
    server = salt.channel.server.ReqServerChannel.factory(master_opts)
    server.auto_key = salt.daemons.masterapi.AutoKey(server.opts)
    server.cache_cli = False
    server.event = salt.utils.event.get_master_event(
        master_opts, master_opts["sock_dir"], listen=False
    )
    server.master_key = salt.crypt.MasterKeys(server.opts)
    minion_opts["verify_master_pubkey_sign"] = True
    minion_opts["always_verify_signature"] = True
    minion_opts["master_sign_key_name"] = "master_sign"
    minion_opts["master"] = "master"

    assert (
        pki_dir.joinpath("minion", "minion_master.pub").read_text()
        == pki_dir.joinpath("master", "master.pub").read_text()
    )

    client = salt.channel.client.AsyncReqChannel.factory(minion_opts, io_loop=io_loop)
    signin_payload = client.auth.minion_sign_in_payload()
    pload = client._package_load(signin_payload)
    assert "version" in pload
    assert pload["version"] == 2

    server_reply = server._auth(pload["load"], sign_messages=True)
    # With version 2 we always get a clear signed response
    assert "enc" in server_reply
    assert server_reply["enc"] == "clear"
    assert "sig" in server_reply
    assert "load" in server_reply
    ret = client.auth.handle_signin_response(signin_payload, server_reply)
    assert "aes" in ret
    assert "master_uri" in ret
    assert "publish_port" in ret

    # Now create a new master key pair and try auth with it.
    mapriv = pki_dir.joinpath("master", "master.pem")
    mapriv.unlink()
    mapriv.write_text(MASTER2_PRIV_KEY.strip())
    mapub = pki_dir.joinpath("master", "master.pub")
    mapub.unlink()
    mapub.write_text(MASTER2_PUB_KEY.strip())

    server = salt.channel.server.ReqServerChannel.factory(master_opts)
    server.auto_key = salt.daemons.masterapi.AutoKey(server.opts)
    server.cache_cli = False
    server.event = salt.utils.event.get_master_event(
        master_opts, master_opts["sock_dir"], listen=False
    )
    server.master_key = salt.crypt.MasterKeys(server.opts)

    signin_payload = client.auth.minion_sign_in_payload()
    pload = client._package_load(signin_payload)
    server_reply = server._auth(pload["load"], sign_messages=True)
    try:
        ret = client.auth.handle_signin_response(signin_payload, server_reply)

        assert "aes" in ret
        assert "master_uri" in ret
        assert "publish_port" in ret

        assert (
            pki_dir.joinpath("minion", "minion_master.pub").read_text()
            == pki_dir.joinpath("master", "master.pub").read_text()
        )
    finally:
        client.close()
        server.close()


async def test_req_chan_auth_v2_new_minion_with_master_pub(
    minion_opts, master_opts, pki_dir, io_loop
):

    pki_dir.joinpath("master", "minions", "minion").unlink()
    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "max_minions": 0,
            "auto_accept": False,
            "open_mode": False,
            "key_pass": None,
            "publish_port": 4505,
            "auth_mode": 1,
            "acceptance_wait_time": 3,
            "acceptance_wait_time_max": 3,
        }
    )
    SMaster.secrets["aes"] = {
        "secret": multiprocessing.Array(
            ctypes.c_char,
            salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
        ),
        "reload": salt.crypt.Crypticle.generate_key_string,
    }
    master_opts.update(pki_dir=str(pki_dir.joinpath("master")))
    master_opts["master_sign_pubkey"] = False
    server = salt.channel.server.ReqServerChannel.factory(master_opts)
    server.auto_key = salt.daemons.masterapi.AutoKey(server.opts)
    server.cache_cli = False
    server.master_key = salt.crypt.MasterKeys(server.opts)
    minion_opts["verify_master_pubkey_sign"] = False
    minion_opts["always_verify_signature"] = False
    client = salt.channel.client.AsyncReqChannel.factory(minion_opts, io_loop=io_loop)
    signin_payload = client.auth.minion_sign_in_payload()
    pload = client._package_load(signin_payload)
    try:
        assert "version" in pload
        assert pload["version"] == 2

        ret = server._auth(pload["load"], sign_messages=True)
        assert "sig" in ret
        ret = client.auth.handle_signin_response(signin_payload, ret)
        assert ret == "retry"
    finally:
        client.close()
        server.close()


async def test_req_chan_auth_v2_new_minion_with_master_pub_bad_sig(
    minion_opts, master_opts, pki_dir, io_loop
):

    pki_dir.joinpath("master", "minions", "minion").unlink()

    # Give the master a different key than the minion has.
    mapriv = pki_dir.joinpath("master", "master.pem")
    mapriv.unlink()
    mapriv.write_text(MASTER2_PRIV_KEY.strip())
    mapub = pki_dir.joinpath("master", "master.pub")
    mapub.unlink()
    mapub.write_text(MASTER2_PUB_KEY.strip())

    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "max_minions": 0,
            "auto_accept": False,
            "open_mode": False,
            "key_pass": None,
            "publish_port": 4505,
            "auth_mode": 1,
            "acceptance_wait_time": 3,
            "acceptance_wait_time_max": 3,
        }
    )
    SMaster.secrets["aes"] = {
        "secret": multiprocessing.Array(
            ctypes.c_char,
            salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
        ),
        "reload": salt.crypt.Crypticle.generate_key_string,
    }
    master_opts.update(pki_dir=str(pki_dir.joinpath("master")))
    master_opts["master_sign_pubkey"] = False
    server = salt.channel.server.ReqServerChannel.factory(master_opts)
    server.auto_key = salt.daemons.masterapi.AutoKey(server.opts)
    server.cache_cli = False
    server.master_key = salt.crypt.MasterKeys(server.opts)
    minion_opts["verify_master_pubkey_sign"] = False
    minion_opts["always_verify_signature"] = False
    client = salt.channel.client.AsyncReqChannel.factory(minion_opts, io_loop=io_loop)
    signin_payload = client.auth.minion_sign_in_payload()
    pload = client._package_load(signin_payload)
    try:
        assert "version" in pload
        assert pload["version"] == 2

        ret = server._auth(pload["load"], sign_messages=True)
        assert "sig" in ret
        with pytest.raises(salt.crypt.SaltClientError, match="Invalid signature"):
            ret = client.auth.handle_signin_response(signin_payload, ret)
    finally:
        client.close()
        server.close()


async def test_req_chan_auth_v2_new_minion_without_master_pub(
    minion_opts, master_opts, pki_dir, io_loop
):

    pki_dir.joinpath("master", "minions", "minion").unlink()
    pki_dir.joinpath("minion", "minion_master.pub").unlink()
    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "max_minions": 0,
            "auto_accept": False,
            "open_mode": False,
            "key_pass": None,
            "publish_port": 4505,
            "auth_mode": 1,
            "acceptance_wait_time": 3,
            "acceptance_wait_time_max": 3,
        }
    )
    SMaster.secrets["aes"] = {
        "secret": multiprocessing.Array(
            ctypes.c_char,
            salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
        ),
        "reload": salt.crypt.Crypticle.generate_key_string,
    }
    master_opts.update(pki_dir=str(pki_dir.joinpath("master")))
    master_opts["master_sign_pubkey"] = False
    server = salt.channel.server.ReqServerChannel.factory(master_opts)
    server.auto_key = salt.daemons.masterapi.AutoKey(server.opts)
    server.cache_cli = False
    server.master_key = salt.crypt.MasterKeys(server.opts)
    minion_opts["verify_master_pubkey_sign"] = False
    minion_opts["always_verify_signature"] = False
    client = salt.channel.client.AsyncReqChannel.factory(minion_opts, io_loop=io_loop)
    signin_payload = client.auth.minion_sign_in_payload()
    pload = client._package_load(signin_payload)
    try:
        assert "version" in pload
        assert pload["version"] == 2

        ret = server._auth(pload["load"], sign_messages=True)
        assert "sig" in ret
        ret = client.auth.handle_signin_response(signin_payload, ret)
        assert ret == "retry"
    finally:
        client.close()
        server.close()


async def test_req_chan_bad_payload_to_decode(
    minion_opts, master_opts, pki_dir, io_loop
):
    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "max_minions": 0,
            "auto_accept": False,
            "open_mode": False,
            "key_pass": None,
            "publish_port": 4505,
            "auth_mode": 1,
            "acceptance_wait_time": 3,
            "acceptance_wait_time_max": 3,
        }
    )
    SMaster.secrets["aes"] = {
        "secret": multiprocessing.Array(
            ctypes.c_char,
            salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
        ),
        "reload": salt.crypt.Crypticle.generate_key_string,
    }
    master_opts.update(dict(minion_opts, pki_dir=str(pki_dir.joinpath("master"))))
    master_opts["master_sign_pubkey"] = False
    server = salt.channel.server.ReqServerChannel.factory(master_opts)

    with pytest.raises(salt.exceptions.SaltDeserializationError):
        server._decode_payload(None)
    with pytest.raises(salt.exceptions.SaltDeserializationError):
        server._decode_payload({})
    with pytest.raises(salt.exceptions.SaltDeserializationError):
        server._decode_payload(12345)


def test_req_server_auth_garbage_sig_algo(pki_dir, minion_opts, master_opts, caplog):
    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "max_minions": 0,
            "auto_accept": False,
            "open_mode": False,
            "key_pass": None,
            "master_sign_pubkey": False,
            "publish_port": 4505,
            "auth_mode": 1,
        }
    )
    SMaster.secrets["aes"] = {
        "secret": multiprocessing.Array(
            ctypes.c_char,
            salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
        ),
        "reload": salt.crypt.Crypticle.generate_key_string,
    }
    master_opts.update(pki_dir=str(pki_dir.joinpath("master")))
    server = salt.channel.server.ReqServerChannel.factory(master_opts)

    server.auto_key = salt.daemons.masterapi.AutoKey(server.opts)
    server.cache_cli = False
    server.event = salt.utils.event.get_master_event(
        master_opts, master_opts["sock_dir"], listen=False
    )
    server.master_key = salt.crypt.MasterKeys(server.opts)
    pub = salt.crypt.PublicKey(str(pki_dir.joinpath("master", "master.pub")))
    token = pub.encrypt(
        salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
        algorithm=minion_opts["encryption_algorithm"],
    )
    nonce = uuid.uuid4().hex

    # We need to read the public key with fopen otherwise the newlines might
    # not match on windows.
    with salt.utils.files.fopen(
        str(pki_dir.joinpath("minion", "minion.pub")), "r"
    ) as fp:
        pub_key = salt.crypt.clean_key(fp.read())

    load = {
        "version": 2,
        "cmd": "_auth",
        "id": "minion",
        "token": token,
        "pub": pub_key,
        "nonce": "asdfse",
        "enc_algo": minion_opts["encryption_algorithm"],
        "sig_algo": "IAMNOTANALGO",
    }
    with caplog.at_level(logging.INFO):
        ret = server._auth(load, sign_messages=True)
        assert (
            "Minion tried to authenticate with unsupported signing algorithm: IAMNOTANALGO"
            in caplog.text
        )
        assert "load" in ret
        assert "ret" in ret["load"]
        assert ret["load"]["ret"] == "bad sig algo"


@pytest.mark.skipif(not FIPS_TESTRUN, reason="Only run on fips enabled platforms")
def test_req_server_auth_unsupported_enc_algo(
    pki_dir, minion_opts, master_opts, caplog
):
    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "max_minions": 0,
            "auto_accept": False,
            "open_mode": False,
            "key_pass": None,
            "master_sign_pubkey": False,
            "publish_port": 4505,
            "auth_mode": 1,
        }
    )
    SMaster.secrets["aes"] = {
        "secret": multiprocessing.Array(
            ctypes.c_char,
            salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
        ),
        "reload": salt.crypt.Crypticle.generate_key_string,
    }
    master_opts.update(pki_dir=str(pki_dir.joinpath("master")))
    server = salt.channel.server.ReqServerChannel.factory(master_opts)

    server.auto_key = salt.daemons.masterapi.AutoKey(server.opts)
    server.cache_cli = False
    server.event = salt.utils.event.get_master_event(
        master_opts, master_opts["sock_dir"], listen=False
    )
    server.master_key = salt.crypt.MasterKeys(server.opts)
    import tests.pytests.unit.crypt

    pub = tests.pytests.unit.crypt.LegacyPublicKey(
        str(pki_dir.joinpath("master", "master.pub"))
    )
    token = pub.encrypt(
        salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
    )
    nonce = uuid.uuid4().hex

    # We need to read the public key with fopen otherwise the newlines might
    # not match on windows.
    with salt.utils.files.fopen(
        str(pki_dir.joinpath("minion", "minion.pub")), "r"
    ) as fp:
        pub_key = salt.crypt.clean_key(fp.read())

    load = {
        "version": 2,
        "cmd": "_auth",
        "id": "minion",
        "token": token,
        "pub": pub_key,
        "nonce": "asdfse",
        "enc_algo": "OAEP-SHA1",
        "sig_algo": minion_opts["signing_algorithm"],
    }
    with caplog.at_level(logging.INFO):
        ret = server._auth(load, sign_messages=True)
        assert (
            "Minion minion tried to authenticate with unsupported encryption algorithm: OAEP-SHA1"
            in caplog.text
        )
        assert "load" in ret
        assert "ret" in ret["load"]
        assert ret["load"]["ret"] == "bad enc algo"


def test_req_server_auth_garbage_enc_algo(pki_dir, minion_opts, master_opts, caplog):
    minion_opts.update(
        {
            "master_uri": "tcp://127.0.0.1:4506",
            "interface": "127.0.0.1",
            "ret_port": 4506,
            "ipv6": False,
            "sock_dir": ".",
            "pki_dir": str(pki_dir.joinpath("minion")),
            "id": "minion",
            "__role": "minion",
            "keysize": 4096,
            "max_minions": 0,
            "auto_accept": False,
            "open_mode": False,
            "key_pass": None,
            "master_sign_pubkey": False,
            "publish_port": 4505,
            "auth_mode": 1,
        }
    )
    SMaster.secrets["aes"] = {
        "secret": multiprocessing.Array(
            ctypes.c_char,
            salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
        ),
        "reload": salt.crypt.Crypticle.generate_key_string,
    }
    master_opts.update(pki_dir=str(pki_dir.joinpath("master")))
    server = salt.channel.server.ReqServerChannel.factory(master_opts)

    server.auto_key = salt.daemons.masterapi.AutoKey(server.opts)
    server.cache_cli = False
    server.event = salt.utils.event.get_master_event(
        master_opts, master_opts["sock_dir"], listen=False
    )
    server.master_key = salt.crypt.MasterKeys(server.opts)
    import tests.pytests.unit.crypt

    pub = tests.pytests.unit.crypt.LegacyPublicKey(
        str(pki_dir.joinpath("master", "master.pub"))
    )
    token = pub.encrypt(
        salt.utils.stringutils.to_bytes(salt.crypt.Crypticle.generate_key_string()),
    )
    nonce = uuid.uuid4().hex

    # We need to read the public key with fopen otherwise the newlines might
    # not match on windows.
    with salt.utils.files.fopen(
        str(pki_dir.joinpath("minion", "minion.pub")), "r"
    ) as fp:
        pub_key = salt.crypt.clean_key(fp.read())

    load = {
        "version": 2,
        "cmd": "_auth",
        "id": "minion",
        "token": token,
        "pub": pub_key,
        "nonce": "asdfse",
        "enc_algo": "IAMNOTAENCALGO",
        "sig_algo": minion_opts["signing_algorithm"],
    }
    with caplog.at_level(logging.INFO):
        ret = server._auth(load, sign_messages=True)
        assert (
            "Minion minion tried to authenticate with unsupported encryption algorithm: IAMNOTAENCALGO"
            in caplog.text
        )
        assert "load" in ret
        assert "ret" in ret["load"]
        assert ret["load"]["ret"] == "bad enc algo"



"""
tests for host state
"""

import pytest

from tests.support.case import ModuleCase


class HandleErrorTest(ModuleCase):
    """
    Validate that ordering works correctly
    """

    @pytest.mark.slow_test
    def test_function_do_not_return_dictionary_type(self):
        """
        Handling a case when function returns anything but a dictionary type
        """
        ret = self.run_function("state.sls", ["issue-9983-handleerror"])
        self.assertTrue(
            "Data must be a dictionary type" in ret[[a for a in ret][0]]["comment"]
        )
        self.assertTrue(not ret[[a for a in ret][0]]["result"])
        self.assertTrue(ret[[a for a in ret][0]]["changes"] == {})



"""
Unit tests for the splunk returner
"""

import json

import pytest

import salt.modules.config as config
import salt.returners.splunk as splunk
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    opts = {
        "splunk_http_forwarder": {
            "token": "TheToken",
            "indexer": "the.splunk.domain",
            "index": "TheIndex",
            "sourcetype": "TheSourceType",
        }
    }
    return {
        splunk: {"__opts__": opts, "__salt__": {"config.get": config.get}},
        config: {"__opts__": opts},
    }


def test_verify_ssl_defaults_to_true():
    payload = {"some": "payload"}
    requests_post = MagicMock()
    ts = 1234565789
    host = "TheHostName"
    data = {
        "time": str(ts),
        "index": "TheIndex",
        "sourcetype": "TheSourceType",
        "event": payload,
        "host": host,
    }
    with patch("salt.returners.splunk.time.time", MagicMock(return_value=ts)), patch(
        "salt.returners.splunk.socket.gethostname", MagicMock(return_value=host)
    ), patch("requests.post", requests_post):
        splunk.returner(payload.copy())
    assert json.loads(requests_post.call_args_list[0][1]["data"]) == data
    assert requests_post.call_args_list[0][1]["verify"]
    assert requests_post.call_args_list[0][1]["headers"] == {
        "Authorization": "Splunk TheToken"
    }
    assert (
        requests_post.call_args_list[0][0][0]
        == "https://the.splunk.domain:8088/services/collector/event"
    )


def test_verify_ssl():
    payload = {"some": "payload"}
    verify_ssl_values = [True, False, None]
    payload = {"some": "payload"}
    ts = 1234565789
    host = "TheHostName"
    data = {
        "time": str(ts),
        "index": "TheIndex",
        "sourcetype": "TheSourceType",
        "event": payload,
        "host": host,
    }
    for verify_ssl in verify_ssl_values:
        requests_post = MagicMock()
        with patch(
            "salt.returners.splunk.time.time", MagicMock(return_value=ts)
        ), patch(
            "salt.returners.splunk.socket.gethostname", MagicMock(return_value=host)
        ), patch(
            "requests.post", requests_post
        ), patch.dict(
            splunk.__opts__["splunk_http_forwarder"], verify_ssl=verify_ssl
        ):
            splunk.returner(payload.copy())
            assert json.loads(requests_post.call_args_list[0][1]["data"]) == data
            assert requests_post.call_args_list[0][1]["verify"] == verify_ssl
            assert requests_post.call_args_list[0][1]["headers"] == {
                "Authorization": "Splunk TheToken"
            }
            assert (
                requests_post.call_args_list[0][0][0]
                == "https://the.splunk.domain:8088/services/collector/event"
            )


def test_verify_event_returner():
    payload = [{"some": "payload"}, {"another": "event"}]
    ts = 1234565789
    host = "TheHostName"
    verify_ssl = True

    requests_post = MagicMock()
    with patch("salt.returners.splunk.time.time", MagicMock(return_value=ts)), patch(
        "salt.returners.splunk.socket.gethostname", MagicMock(return_value=host)
    ), patch("requests.post", requests_post), patch.dict(
        splunk.__opts__["splunk_http_forwarder"], verify_ssl=verify_ssl
    ):
        splunk.event_return(payload)
        for i in range(len(payload)):
            assert (
                json.loads(requests_post.call_args_list[0][1]["data"])["event"]
                in payload
            )
            assert requests_post.call_args_list[0][1]["verify"] == verify_ssl
            assert requests_post.call_args_list[0][1]["headers"] == {
                "Authorization": "Splunk TheToken"
            }
            assert (
                requests_post.call_args_list[0][0][0]
                == "https://the.splunk.domain:8088/services/collector/event"
            )



import salt.channel.client


def test_async_methods():
    "Validate all defined async_methods and close_methods are present"
    async_classes = [
        salt.channel.client.AsyncReqChannel,
        salt.channel.client.AsyncPubChannel,
    ]
    method_attrs = [
        "async_methods",
        "close_methods",
    ]
    for cls in async_classes:
        for attr in method_attrs:
            assert hasattr(cls, attr)
            assert isinstance(getattr(cls, attr), list)
            for name in getattr(cls, attr):
                assert hasattr(cls, name)



"""
Manage kubernetes resources as salt states
==========================================

NOTE: This module requires the proper pillar values set. See
salt.modules.kubernetesmod for more information.

.. warning::

    Configuration options will change in 2019.2.0.

The kubernetes module is used to manage different kubernetes resources.


.. code-block:: yaml

    my-nginx:
      kubernetes.deployment_present:
        - namespace: default
          metadata:
            app: frontend
          spec:
            replicas: 1
            template:
              metadata:
                labels:
                  run: my-nginx
              spec:
                containers:
                - name: my-nginx
                  image: nginx
                  ports:
                  - containerPort: 80

    my-mariadb:
      kubernetes.deployment_absent:
        - namespace: default

    # kubernetes deployment as specified inside of
    # a file containing the definition of the the
    # deployment using the official kubernetes format
    redis-master-deployment:
      kubernetes.deployment_present:
        - name: redis-master
        - source: salt://k8s/redis-master-deployment.yml
      require:
        - pip: kubernetes-python-module

    # kubernetes service as specified inside of
    # a file containing the definition of the the
    # service using the official kubernetes format
    redis-master-service:
      kubernetes.service_present:
        - name: redis-master
        - source: salt://k8s/redis-master-service.yml
      require:
        - kubernetes.deployment_present: redis-master

    # kubernetes deployment as specified inside of
    # a file containing the definition of the the
    # deployment using the official kubernetes format
    # plus some jinja directives
     nginx-source-template:
      kubernetes.deployment_present:
        - source: salt://k8s/nginx.yml.jinja
        - template: jinja
      require:
        - pip: kubernetes-python-module


    # Kubernetes secret
    k8s-secret:
      kubernetes.secret_present:
        - name: top-secret
          data:
            key1: value1
            key2: value2
            key3: value3

.. versionadded:: 2017.7.0
"""

import copy
import logging

log = logging.getLogger(__name__)

__deprecated__ = (
    3009,
    "kubernetes",
    "https://github.com/salt-extensions/saltext-kubernetes",
)


def __virtual__():
    """
    Only load if the kubernetes module is available in __salt__
    """
    if "kubernetes.ping" in __salt__:
        return True
    return (False, "kubernetes module could not be loaded")


def _error(ret, err_msg):
    """
    Helper function to propagate errors to
    the end user.
    """
    ret["result"] = False
    ret["comment"] = err_msg
    return ret


def deployment_absent(name, namespace="default", **kwargs):
    """
    Ensures that the named deployment is absent from the given namespace.

    name
        The name of the deployment

    namespace
        The name of the namespace
    """

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    deployment = __salt__["kubernetes.show_deployment"](name, namespace, **kwargs)

    if deployment is None:
        ret["result"] = True if not __opts__["test"] else None
        ret["comment"] = "The deployment does not exist"
        return ret

    if __opts__["test"]:
        ret["comment"] = "The deployment is going to be deleted"
        ret["result"] = None
        return ret

    res = __salt__["kubernetes.delete_deployment"](name, namespace, **kwargs)
    if res["code"] == 200:
        ret["result"] = True
        ret["changes"] = {"kubernetes.deployment": {"new": "absent", "old": "present"}}
        ret["comment"] = res["message"]
    else:
        ret["comment"] = f"Something went wrong, response: {res}"

    return ret


def deployment_present(
    name,
    namespace="default",
    metadata=None,
    spec=None,
    source="",
    template="",
    **kwargs,
):
    """
    Ensures that the named deployment is present inside of the specified
    namespace with the given metadata and spec.
    If the deployment exists it will be replaced.

    name
        The name of the deployment.

    namespace
        The namespace holding the deployment. The 'default' one is going to be
        used unless a different one is specified.

    metadata
        The metadata of the deployment object.

    spec
        The spec of the deployment object.

    source
        A file containing the definition of the deployment (metadata and
        spec) in the official kubernetes format.

    template
        Template engine to be used to render the source file.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if (metadata or spec) and source:
        return _error(
            ret, "'source' cannot be used in combination with 'metadata' or 'spec'"
        )

    if metadata is None:
        metadata = {}

    if spec is None:
        spec = {}

    deployment = __salt__["kubernetes.show_deployment"](name, namespace, **kwargs)

    if deployment is None:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "The deployment is going to be created"
            return ret
        res = __salt__["kubernetes.create_deployment"](
            name=name,
            namespace=namespace,
            metadata=metadata,
            spec=spec,
            source=source,
            template=template,
            saltenv=__env__,
            **kwargs,
        )
        ret["changes"][f"{namespace}.{name}"] = {"old": {}, "new": res}
    else:
        if __opts__["test"]:
            ret["result"] = None
            return ret

        # TODO: improve checks  # pylint: disable=fixme
        log.info("Forcing the recreation of the deployment")
        ret["comment"] = "The deployment is already present. Forcing recreation"
        res = __salt__["kubernetes.replace_deployment"](
            name=name,
            namespace=namespace,
            metadata=metadata,
            spec=spec,
            source=source,
            template=template,
            saltenv=__env__,
            **kwargs,
        )

    ret["changes"] = {"metadata": metadata, "spec": spec}
    ret["result"] = True
    return ret


def service_present(
    name,
    namespace="default",
    metadata=None,
    spec=None,
    source="",
    template="",
    **kwargs,
):
    """
    Ensures that the named service is present inside of the specified namespace
    with the given metadata and spec.
    If the deployment exists it will be replaced.

    name
        The name of the service.

    namespace
        The namespace holding the service. The 'default' one is going to be
        used unless a different one is specified.

    metadata
        The metadata of the service object.

    spec
        The spec of the service object.

    source
        A file containing the definition of the service (metadata and
        spec) in the official kubernetes format.

    template
        Template engine to be used to render the source file.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if (metadata or spec) and source:
        return _error(
            ret, "'source' cannot be used in combination with 'metadata' or 'spec'"
        )

    if metadata is None:
        metadata = {}

    if spec is None:
        spec = {}

    service = __salt__["kubernetes.show_service"](name, namespace, **kwargs)

    if service is None:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "The service is going to be created"
            return ret
        res = __salt__["kubernetes.create_service"](
            name=name,
            namespace=namespace,
            metadata=metadata,
            spec=spec,
            source=source,
            template=template,
            saltenv=__env__,
            **kwargs,
        )
        ret["changes"][f"{namespace}.{name}"] = {"old": {}, "new": res}
    else:
        if __opts__["test"]:
            ret["result"] = None
            return ret

        # TODO: improve checks  # pylint: disable=fixme
        log.info("Forcing the recreation of the service")
        ret["comment"] = "The service is already present. Forcing recreation"
        res = __salt__["kubernetes.replace_service"](
            name=name,
            namespace=namespace,
            metadata=metadata,
            spec=spec,
            source=source,
            template=template,
            old_service=service,
            saltenv=__env__,
            **kwargs,
        )

    ret["changes"] = {"metadata": metadata, "spec": spec}
    ret["result"] = True
    return ret


def service_absent(name, namespace="default", **kwargs):
    """
    Ensures that the named service is absent from the given namespace.

    name
        The name of the service

    namespace
        The name of the namespace
    """

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    service = __salt__["kubernetes.show_service"](name, namespace, **kwargs)

    if service is None:
        ret["result"] = True if not __opts__["test"] else None
        ret["comment"] = "The service does not exist"
        return ret

    if __opts__["test"]:
        ret["comment"] = "The service is going to be deleted"
        ret["result"] = None
        return ret

    res = __salt__["kubernetes.delete_service"](name, namespace, **kwargs)
    if res["code"] == 200:
        ret["result"] = True
        ret["changes"] = {"kubernetes.service": {"new": "absent", "old": "present"}}
        ret["comment"] = res["message"]
    else:
        ret["comment"] = f"Something went wrong, response: {res}"

    return ret


def namespace_absent(name, **kwargs):
    """
    Ensures that the named namespace is absent.

    name
        The name of the namespace
    """

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    namespace = __salt__["kubernetes.show_namespace"](name, **kwargs)

    if namespace is None:
        ret["result"] = True if not __opts__["test"] else None
        ret["comment"] = "The namespace does not exist"
        return ret

    if __opts__["test"]:
        ret["comment"] = "The namespace is going to be deleted"
        ret["result"] = None
        return ret

    res = __salt__["kubernetes.delete_namespace"](name, **kwargs)
    if (
        res["code"] == 200
        or (isinstance(res["status"], str) and "Terminating" in res["status"])
        or (isinstance(res["status"], dict) and res["status"]["phase"] == "Terminating")
    ):
        ret["result"] = True
        ret["changes"] = {"kubernetes.namespace": {"new": "absent", "old": "present"}}
        if res["message"]:
            ret["comment"] = res["message"]
        else:
            ret["comment"] = "Terminating"
    else:
        ret["comment"] = f"Something went wrong, response: {res}"

    return ret


def namespace_present(name, **kwargs):
    """
    Ensures that the named namespace is present.

    name
        The name of the namespace.

    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    namespace = __salt__["kubernetes.show_namespace"](name, **kwargs)

    if namespace is None:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "The namespace is going to be created"
            return ret

        res = __salt__["kubernetes.create_namespace"](name, **kwargs)
        ret["result"] = True
        ret["changes"]["namespace"] = {"old": {}, "new": res}
    else:
        ret["result"] = True if not __opts__["test"] else None
        ret["comment"] = "The namespace already exists"

    return ret


def secret_absent(name, namespace="default", **kwargs):
    """
    Ensures that the named secret is absent from the given namespace.

    name
        The name of the secret

    namespace
        The name of the namespace
    """

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    secret = __salt__["kubernetes.show_secret"](name, namespace, **kwargs)

    if secret is None:
        ret["result"] = True if not __opts__["test"] else None
        ret["comment"] = "The secret does not exist"
        return ret

    if __opts__["test"]:
        ret["comment"] = "The secret is going to be deleted"
        ret["result"] = None
        return ret

    __salt__["kubernetes.delete_secret"](name, namespace, **kwargs)

    # As for kubernetes 1.6.4 doesn't set a code when deleting a secret
    # The kubernetes module will raise an exception if the kubernetes
    # server will return an error
    ret["result"] = True
    ret["changes"] = {"kubernetes.secret": {"new": "absent", "old": "present"}}
    ret["comment"] = "Secret deleted"
    return ret


def secret_present(
    name, namespace="default", data=None, source=None, template=None, **kwargs
):
    """
    Ensures that the named secret is present inside of the specified namespace
    with the given data.
    If the secret exists it will be replaced.

    name
        The name of the secret.

    namespace
        The namespace holding the secret. The 'default' one is going to be
        used unless a different one is specified.

    data
        The dictionary holding the secrets.

    source
        A file containing the data of the secret in plain format.

    template
        Template engine to be used to render the source file.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if data and source:
        return _error(ret, "'source' cannot be used in combination with 'data'")

    secret = __salt__["kubernetes.show_secret"](name, namespace, **kwargs)

    if secret is None:
        if data is None:
            data = {}

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "The secret is going to be created"
            return ret
        res = __salt__["kubernetes.create_secret"](
            name=name,
            namespace=namespace,
            data=data,
            source=source,
            template=template,
            saltenv=__env__,
            **kwargs,
        )
        ret["changes"][f"{namespace}.{name}"] = {"old": {}, "new": res}
    else:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "The secret is going to be replaced"
            return ret

        # TODO: improve checks  # pylint: disable=fixme
        log.info("Forcing the recreation of the service")
        ret["comment"] = "The secret is already present. Forcing recreation"
        res = __salt__["kubernetes.replace_secret"](
            name=name,
            namespace=namespace,
            data=data,
            source=source,
            template=template,
            saltenv=__env__,
            **kwargs,
        )

    ret["changes"] = {
        # Omit values from the return. They are unencrypted
        # and can contain sensitive data.
        "data": list(res["data"])
    }
    ret["result"] = True

    return ret


def configmap_absent(name, namespace="default", **kwargs):
    """
    Ensures that the named configmap is absent from the given namespace.

    name
        The name of the configmap

    namespace
        The namespace holding the configmap. The 'default' one is going to be
        used unless a different one is specified.
    """

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    configmap = __salt__["kubernetes.show_configmap"](name, namespace, **kwargs)

    if configmap is None:
        ret["result"] = True if not __opts__["test"] else None
        ret["comment"] = "The configmap does not exist"
        return ret

    if __opts__["test"]:
        ret["comment"] = "The configmap is going to be deleted"
        ret["result"] = None
        return ret

    __salt__["kubernetes.delete_configmap"](name, namespace, **kwargs)
    # As for kubernetes 1.6.4 doesn't set a code when deleting a configmap
    # The kubernetes module will raise an exception if the kubernetes
    # server will return an error
    ret["result"] = True
    ret["changes"] = {"kubernetes.configmap": {"new": "absent", "old": "present"}}
    ret["comment"] = "ConfigMap deleted"

    return ret


def configmap_present(
    name, namespace="default", data=None, source=None, template=None, **kwargs
):
    """
    Ensures that the named configmap is present inside of the specified namespace
    with the given data.
    If the configmap exists it will be replaced.

    name
        The name of the configmap.

    namespace
        The namespace holding the configmap. The 'default' one is going to be
        used unless a different one is specified.

    data
        The dictionary holding the configmaps.

    source
        A file containing the data of the configmap in plain format.

    template
        Template engine to be used to render the source file.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if data and source:
        return _error(ret, "'source' cannot be used in combination with 'data'")
    elif data is None:
        data = {}

    configmap = __salt__["kubernetes.show_configmap"](name, namespace, **kwargs)

    if configmap is None:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "The configmap is going to be created"
            return ret
        res = __salt__["kubernetes.create_configmap"](
            name=name,
            namespace=namespace,
            data=data,
            source=source,
            template=template,
            saltenv=__env__,
            **kwargs,
        )
        ret["changes"][f"{namespace}.{name}"] = {"old": {}, "new": res}
    else:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "The configmap is going to be replaced"
            return ret

        # TODO: improve checks  # pylint: disable=fixme
        log.info("Forcing the recreation of the service")
        ret["comment"] = "The configmap is already present. Forcing recreation"
        res = __salt__["kubernetes.replace_configmap"](
            name=name,
            namespace=namespace,
            data=data,
            source=source,
            template=template,
            saltenv=__env__,
            **kwargs,
        )

    ret["changes"] = {"data": res["data"]}
    ret["result"] = True
    return ret


def pod_absent(name, namespace="default", **kwargs):
    """
    Ensures that the named pod is absent from the given namespace.

    name
        The name of the pod

    namespace
        The name of the namespace
    """

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    pod = __salt__["kubernetes.show_pod"](name, namespace, **kwargs)

    if pod is None:
        ret["result"] = True if not __opts__["test"] else None
        ret["comment"] = "The pod does not exist"
        return ret

    if __opts__["test"]:
        ret["comment"] = "The pod is going to be deleted"
        ret["result"] = None
        return ret

    res = __salt__["kubernetes.delete_pod"](name, namespace, **kwargs)
    if res["code"] == 200 or res["code"] is None:
        ret["result"] = True
        ret["changes"] = {"kubernetes.pod": {"new": "absent", "old": "present"}}
        if res["code"] is None:
            ret["comment"] = "In progress"
        else:
            ret["comment"] = res["message"]
    else:
        ret["comment"] = f"Something went wrong, response: {res}"

    return ret


def pod_present(
    name,
    namespace="default",
    metadata=None,
    spec=None,
    source="",
    template="",
    **kwargs,
):
    """
    Ensures that the named pod is present inside of the specified
    namespace with the given metadata and spec.
    If the pod exists it will be replaced.

    name
        The name of the pod.

    namespace
        The namespace holding the pod. The 'default' one is going to be
        used unless a different one is specified.

    metadata
        The metadata of the pod object.

    spec
        The spec of the pod object.

    source
        A file containing the definition of the pod (metadata and
        spec) in the official kubernetes format.

    template
        Template engine to be used to render the source file.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if (metadata or spec) and source:
        return _error(
            ret, "'source' cannot be used in combination with 'metadata' or 'spec'"
        )

    if metadata is None:
        metadata = {}

    if spec is None:
        spec = {}

    pod = __salt__["kubernetes.show_pod"](name, namespace, **kwargs)

    if pod is None:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "The pod is going to be created"
            return ret
        res = __salt__["kubernetes.create_pod"](
            name=name,
            namespace=namespace,
            metadata=metadata,
            spec=spec,
            source=source,
            template=template,
            saltenv=__env__,
            **kwargs,
        )
        ret["changes"][f"{namespace}.{name}"] = {"old": {}, "new": res}
    else:
        if __opts__["test"]:
            ret["result"] = None
            return ret

        # TODO: fix replace_namespaced_pod validation issues
        ret["comment"] = (
            "salt is currently unable to replace a pod without "
            "deleting it. Please perform the removal of the pod requiring "
            "the 'pod_absent' state if this is the desired behaviour."
        )
        ret["result"] = False
        return ret

    ret["changes"] = {"metadata": metadata, "spec": spec}
    ret["result"] = True
    return ret


def node_label_absent(name, node, **kwargs):
    """
    Ensures that the named label is absent from the node.

    name
        The name of the label

    node
        The name of the node
    """

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    labels = __salt__["kubernetes.node_labels"](node, **kwargs)

    if name not in labels:
        ret["result"] = True if not __opts__["test"] else None
        ret["comment"] = "The label does not exist"
        return ret

    if __opts__["test"]:
        ret["comment"] = "The label is going to be deleted"
        ret["result"] = None
        return ret

    __salt__["kubernetes.node_remove_label"](node_name=node, label_name=name, **kwargs)

    ret["result"] = True
    ret["changes"] = {"kubernetes.node_label": {"new": "absent", "old": "present"}}
    ret["comment"] = "Label removed from node"

    return ret


def node_label_folder_absent(name, node, **kwargs):
    """
    Ensures the label folder doesn't exist on the specified node.

    name
        The name of label folder

    node
        The name of the node
    """

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    labels = __salt__["kubernetes.node_labels"](node, **kwargs)

    folder = name.strip("/") + "/"
    labels_to_drop = []
    new_labels = []
    for label in labels:
        if label.startswith(folder):
            labels_to_drop.append(label)
        else:
            new_labels.append(label)

    if not labels_to_drop:
        ret["result"] = True if not __opts__["test"] else None
        ret["comment"] = "The label folder does not exist"
        return ret

    if __opts__["test"]:
        ret["comment"] = "The label folder is going to be deleted"
        ret["result"] = None
        return ret

    for label in labels_to_drop:
        __salt__["kubernetes.node_remove_label"](
            node_name=node, label_name=label, **kwargs
        )

    ret["result"] = True
    ret["changes"] = {
        "kubernetes.node_label_folder_absent": {"old": list(labels), "new": new_labels}
    }
    ret["comment"] = "Label folder removed from node"

    return ret


def node_label_present(name, node, value, **kwargs):
    """
    Ensures that the named label is set on the named node
    with the given value.
    If the label exists it will be replaced.

    name
        The name of the label.

    value
        Value of the label.

    node
        Node to change.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    labels = __salt__["kubernetes.node_labels"](node, **kwargs)

    if name not in labels:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "The label is going to be set"
            return ret
        __salt__["kubernetes.node_add_label"](
            label_name=name, label_value=value, node_name=node, **kwargs
        )
    elif labels[name] == value:
        ret["result"] = True
        ret["comment"] = "The label is already set and has the specified value"
        return ret
    else:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "The label is going to be updated"
            return ret

        ret["comment"] = "The label is already set, changing the value"
        __salt__["kubernetes.node_add_label"](
            node_name=node, label_name=name, label_value=value, **kwargs
        )

    old_labels = copy.copy(labels)
    labels[name] = value

    ret["changes"][f"{node}.{name}"] = {"old": old_labels, "new": labels}
    ret["result"] = True

    return ret



"""
Management of PostgreSQL Privileges
===================================

The postgres_privileges module is used to manage Postgres privileges.
Privileges can be set as either absent or present.

Privileges can be set on the following database object types:

* database
* schema
* tablespace
* table
* sequence
* language
* group

Setting the grant option is supported as well.

.. versionadded:: 2016.3.0

.. code-block:: yaml

    baruwa:
      postgres_privileges.present:
        - object_name: awl
        - object_type: table
        - privileges:
          - SELECT
          - INSERT
          - DELETE
        - grant_option: False
        - prepend: public
        - maintenance_db: testdb

.. code-block:: yaml

    andrew:
      postgres_privileges.present:
        - object_name: admins
        - object_type: group
        - grant_option: False
        - maintenance_db: testdb

.. code-block:: yaml

    baruwa:
      postgres_privileges.absent:
        - object_name: awl
        - object_type: table
        - privileges:
          - SELECT
          - INSERT
          - DELETE
        - prepend: public
        - maintenance_db: testdb

.. code-block:: yaml

    andrew:
      postgres_privileges.absent:
        - object_name: admins
        - object_type: group
        - maintenance_db: testdb
"""


def __virtual__():
    """
    Only load if the postgres module is present
    """
    if "postgres.privileges_grant" not in __salt__:
        return (
            False,
            "Unable to load postgres module.  Make sure `postgres.bins_dir` is set.",
        )
    return True


def present(
    name,
    object_name,
    object_type,
    privileges=None,
    grant_option=None,
    prepend="public",
    maintenance_db=None,
    user=None,
    db_password=None,
    db_host=None,
    db_port=None,
    db_user=None,
):
    """
    Grant the requested privilege(s) on the specified object to a role

    name
        Name of the role to which privileges should be granted

    object_name
       Name of the object on which the grant is to be performed.
       'ALL' may be used for objects of type 'table' or 'sequence'.

    object_type
       The object type, which can be one of the following:

       - table
       - sequence
       - schema
       - tablespace
       - language
       - database
       - group
       - function

       View permissions should specify `object_type: table`.

    privileges
       List of privileges to grant, from the list below:

       - INSERT
       - CREATE
       - TRUNCATE
       - CONNECT
       - TRIGGER
       - SELECT
       - USAGE
       - TEMPORARY
       - UPDATE
       - EXECUTE
       - REFERENCES
       - DELETE
       - MAINTAIN
       - ALL

       :note: privileges should not be set when granting group membership

    grant_option
        If grant_option is set to True, the recipient of the privilege can
        in turn grant it to others

    prepend
        Table and Sequence object types live under a schema so this should be
        provided if the object is not under the default `public` schema

    maintenance_db
        The name of the database in which the language is to be installed

    user
        System user all operations should be performed on behalf of

    db_user
        database username if different from config or default

    db_password
        user password if any password for a specified user

    db_host
        Database host if different from config or default

    db_port
        Database port if different from config or default
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "The requested privilege(s) are already set",
    }

    privileges = ",".join(privileges) if privileges else None

    kwargs = {
        "privileges": privileges,
        "grant_option": grant_option,
        "prepend": prepend,
        "maintenance_db": maintenance_db,
        "runas": user,
        "host": db_host,
        "user": db_user,
        "port": db_port,
        "password": db_password,
    }

    if not __salt__["postgres.has_privileges"](
        name, object_name, object_type, **kwargs
    ):
        _privs = object_name if object_type == "group" else privileges

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "The privilege(s): {} are set to be granted to {}".format(
                _privs, name
            )
            return ret

        if __salt__["postgres.privileges_grant"](
            name, object_name, object_type, **kwargs
        ):
            ret["comment"] = "The privilege(s): {} have been granted to {}".format(
                _privs, name
            )
            ret["changes"][name] = "Present"
        else:
            ret["comment"] = "Failed to grant privilege(s): {} to {}".format(
                _privs, name
            )
            ret["result"] = False

    return ret


def absent(
    name,
    object_name,
    object_type,
    privileges=None,
    prepend="public",
    maintenance_db=None,
    user=None,
    db_password=None,
    db_host=None,
    db_port=None,
    db_user=None,
):
    """
    Revoke the requested privilege(s) on the specificed object(s)

    name
        Name of the role whose privileges should be revoked

    object_name
       Name of the object on which the revoke is to be performed

    object_type
       The object type, which can be one of the following:

       - table
       - sequence
       - schema
       - tablespace
       - language
       - database
       - group
       - function

       View permissions should specify `object_type: table`.

    privileges
       Comma separated list of privileges to revoke, from the list below:

       - INSERT
       - CREATE
       - TRUNCATE
       - CONNECT
       - TRIGGER
       - SELECT
       - USAGE
       - TEMPORARY
       - UPDATE
       - EXECUTE
       - REFERENCES
       - DELETE
       - MAINTAIN
       - ALL

       :note: privileges should not be set when revoking group membership

    prepend
        Table and Sequence object types live under a schema so this should be
        provided if the object is not under the default `public` schema

    maintenance_db
        The name of the database in which the language is to be installed

    user
        System user all operations should be performed on behalf of

    db_user
        database username if different from config or default

    db_password
        user password if any password for a specified user

    db_host
        Database host if different from config or default

    db_port
        Database port if different from config or default
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": "The requested privilege(s) are not set so cannot be revoked",
    }

    privileges = ",".join(privileges) if privileges else None

    kwargs = {
        "privileges": privileges,
        "prepend": prepend,
        "maintenance_db": maintenance_db,
        "runas": user,
        "host": db_host,
        "user": db_user,
        "port": db_port,
        "password": db_password,
    }

    if __salt__["postgres.has_privileges"](name, object_name, object_type, **kwargs):
        _privs = object_name if object_type == "group" else privileges

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = (
                "The privilege(s): {} are set to be revoked from {}".format(
                    _privs, name
                )
            )
            return ret

        if __salt__["postgres.privileges_revoke"](
            name, object_name, object_type, **kwargs
        ):
            ret["comment"] = "The privilege(s): {} have been revoked from {}".format(
                _privs, name
            )
            ret["changes"][name] = "Absent"
        else:
            ret["comment"] = "Failed to revoke privilege(s): {} from {}".format(
                _privs, name
            )
            ret["result"] = False

    return ret



import pytest


@pytest.fixture(scope="module")
def pillar_tree(salt_master, salt_minion):
    top_file = """
    base:
      '{}':
        - sdb
    """.format(
        salt_minion.id
    )
    sdb_pillar_file = """
    test_etcd_pillar_sdb: sdb://sdbetcd/secret/test/test_pillar_sdb/foo
    """
    top_tempfile = salt_master.pillar_tree.base.temp_file("top.sls", top_file)
    sdb_tempfile = salt_master.pillar_tree.base.temp_file("sdb.sls", sdb_pillar_file)

    with top_tempfile, sdb_tempfile:
        yield



"""
Test the win_wusa execution module
"""

import pytest

import salt.modules.win_wusa as win_wusa
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.fixture
def configure_loader_modules():
    return {win_wusa: {}}


def test_is_installed_false():
    """
    test is_installed function when the KB is not installed
    """
    mock_retcode = MagicMock(return_value=1)
    with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
        assert win_wusa.is_installed("KB123456") is False


def test_is_installed_true():
    """
    test is_installed function when the KB is installed
    """
    mock_retcode = MagicMock(return_value=0)
    with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
        assert win_wusa.is_installed("KB123456") is True


def test_list():
    """
    test list function
    """
    ret = [{"HotFixID": "KB123456"}, {"HotFixID": "KB123457"}]
    mock_all = MagicMock(return_value=ret)
    with patch("salt.utils.win_pwsh.run_dict", mock_all):
        expected = ["KB123456", "KB123457"]
        returned = win_wusa.list_()
        assert returned == expected


def test_install():
    """
    test install function
    """
    mock_retcode = MagicMock(return_value=0)
    path = "C:\\KB123456.msu"
    with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
        assert win_wusa.install(path) is True
    mock_retcode.assert_called_once_with(
        ["wusa.exe", path, "/quiet", "/norestart"], ignore_retcode=True
    )


def test_install_restart():
    """
    test install function with restart=True
    """
    mock_retcode = MagicMock(return_value=0)
    path = "C:\\KB123456.msu"
    with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
        assert win_wusa.install(path, restart=True) is True
    mock_retcode.assert_called_once_with(
        ["wusa.exe", path, "/quiet", "/forcerestart"], ignore_retcode=True
    )


def test_install_already_installed():
    """
    test install function when KB already installed
    """
    retcode = 2359302
    mock_retcode = MagicMock(return_value=retcode)
    path = "C:\\KB123456.msu"
    name = "KB123456.msu"
    with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
        with pytest.raises(CommandExecutionError) as excinfo:
            win_wusa.install(path)
    mock_retcode.assert_called_once_with(
        ["wusa.exe", path, "/quiet", "/norestart"], ignore_retcode=True
    )
    assert (
        f"{name} is already installed. Additional info follows:\n\n{retcode}"
        == excinfo.value.message
    )


def test_install_reboot_needed():
    """
    test install function when KB need a reboot
    """
    retcode = 3010
    mock_retcode = MagicMock(return_value=retcode)
    path = "C:\\KB123456.msu"
    name = "KB123456.msu"
    with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
        with pytest.raises(CommandExecutionError) as excinfo:
            win_wusa.install(path)
    mock_retcode.assert_called_once_with(
        ["wusa.exe", path, "/quiet", "/norestart"], ignore_retcode=True
    )
    assert (
        f"{name} correctly installed but server reboot is needed to complete installation. Additional info follows:\n\n{retcode}"
        == excinfo.value.message
    )


def test_install_error_87():
    """
    test install function when error 87 returned
    """
    retcode = 87
    mock_retcode = MagicMock(return_value=retcode)
    path = "C:\\KB123456.msu"
    with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
        with pytest.raises(CommandExecutionError) as excinfo:
            win_wusa.install(path)
    mock_retcode.assert_called_once_with(
        ["wusa.exe", path, "/quiet", "/norestart"], ignore_retcode=True
    )
    assert (
        f"Unknown error. Additional info follows:\n\n{retcode}" == excinfo.value.message
    )


def test_install_error_other():
    """
    test install function on other unknown error
    """
    mock_retcode = MagicMock(return_value=1234)
    path = "C:\\KB123456.msu"
    with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
        with pytest.raises(CommandExecutionError) as excinfo:
            win_wusa.install(path)
    mock_retcode.assert_called_once_with(
        ["wusa.exe", path, "/quiet", "/norestart"], ignore_retcode=True
    )
    assert "Unknown error: 1234" == excinfo.value.message


def test_uninstall_kb():
    """
    test uninstall function passing kb name
    """
    mock_retcode = MagicMock(return_value=0)
    kb = "KB123456"
    with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}), patch(
        "os.path.exists", MagicMock(return_value=False)
    ):
        assert win_wusa.uninstall(kb) is True
    mock_retcode.assert_called_once_with(
        [
            "wusa.exe",
            "/uninstall",
            "/quiet",
            f"/kb:{kb[2:]}",
            "/norestart",
        ],
        ignore_retcode=True,
    )


def test_uninstall_path():
    """
    test uninstall function passing full path to .msu file
    """
    mock_retcode = MagicMock(return_value=0)
    path = "C:\\KB123456.msu"
    with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}), patch(
        "os.path.exists", MagicMock(return_value=True)
    ):
        assert win_wusa.uninstall(path) is True
    mock_retcode.assert_called_once_with(
        ["wusa.exe", "/uninstall", "/quiet", path, "/norestart"],
        ignore_retcode=True,
    )


def test_uninstall_path_restart():
    """
    test uninstall function with full path and restart=True
    """
    mock_retcode = MagicMock(return_value=0)
    path = "C:\\KB123456.msu"
    with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}), patch(
        "os.path.exists", MagicMock(return_value=True)
    ):
        assert win_wusa.uninstall(path, restart=True) is True
    mock_retcode.assert_called_once_with(
        ["wusa.exe", "/uninstall", "/quiet", path, "/forcerestart"],
        ignore_retcode=True,
    )


def test_uninstall_already_uninstalled():
    """
    test uninstall function when KB already uninstalled
    """
    retcode = 2359303
    mock_retcode = MagicMock(return_value=retcode)
    kb = "KB123456"
    with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}):
        with pytest.raises(CommandExecutionError) as excinfo:
            win_wusa.uninstall(kb)
    mock_retcode.assert_called_once_with(
        [
            "wusa.exe",
            "/uninstall",
            "/quiet",
            f"/kb:{kb[2:]}",
            "/norestart",
        ],
        ignore_retcode=True,
    )
    assert (
        f"{kb} not installed. Additional info follows:\n\n{retcode}"
        == excinfo.value.message
    )


def test_uninstall_path_error_other():
    """
    test uninstall function with unknown error
    """
    mock_retcode = MagicMock(return_value=1234)
    path = "C:\\KB123456.msu"
    with patch.dict(win_wusa.__salt__, {"cmd.retcode": mock_retcode}), patch(
        "os.path.exists", MagicMock(return_value=True)
    ), pytest.raises(CommandExecutionError) as excinfo:
        win_wusa.uninstall(path)
    mock_retcode.assert_called_once_with(
        ["wusa.exe", "/uninstall", "/quiet", path, "/norestart"],
        ignore_retcode=True,
    )
    assert "Unknown error: 1234" == excinfo.value.message



import pytest

import salt.utils.vault as vault
import salt.utils.vault.cache as vcache
import salt.utils.vault.client as vclient
import salt.utils.vault.leases as leases
from tests.support.mock import Mock, call, patch


@pytest.fixture(autouse=True, params=[0])
def time_stopped(request):
    with patch(
        "salt.utils.vault.leases.time.time", autospec=True, return_value=request.param
    ):
        yield


@pytest.fixture
def lease_renewed_response():
    return {
        "lease_id": "database/creds/testrole/abcd",
        "renewable": True,
        "lease_duration": 2000,
    }


@pytest.fixture
def lease_renewed_extended_response():
    return {
        "lease_id": "database/creds/testrole/abcd",
        "renewable": True,
        "lease_duration": 3000,
    }


@pytest.fixture
def store(events):
    client = Mock(spec=vclient.AuthenticatedVaultClient)
    cache = Mock(spec=vcache.VaultLeaseCache)
    cache.exists.return_value = False
    cache.get.return_value = None
    return leases.LeaseStore(client, cache, expire_events=events)


@pytest.fixture
def store_valid(store, lease, lease_renewed_response):
    store.cache.exists.return_value = True
    store.cache.get.return_value = leases.VaultLease(**lease)
    store.client.post.return_value = lease_renewed_response
    return store


@pytest.mark.parametrize(
    "creation_time",
    [
        1661188581,
        "1661188581",
        "2022-08-22T17:16:21.473219641+00:00",
        "2022-08-22T17:16:21.47321964+00:00",
        "2022-08-22T17:16:21.4732196+00:00",
        "2022-08-22T17:16:21.473219+00:00",
        "2022-08-22T17:16:21.47321+00:00",
        "2022-08-22T17:16:21.4732+00:00",
        "2022-08-22T17:16:21.473+00:00",
        "2022-08-22T17:16:21.47+00:00",
        "2022-08-22T17:16:21.4+00:00",
    ],
)
def test_vault_lease_creation_time_normalization(creation_time):
    """
    Ensure the normalization of different creation_time formats works as expected -
    many token endpoints report a timestamp, while other endpoints report RFC3339-formatted
    strings that may have a variable number of digits for sub-second precision (0 omitted)
    while datetime.fromisoformat expects exactly 6 digits
    """
    data = {
        "lease_id": "id",
        "renewable": False,
        "lease_duration": 1337,
        "creation_time": creation_time,
        "data": None,
    }
    res = leases.VaultLease(**data)
    assert res.creation_time == 1661188581


@pytest.mark.parametrize(
    "time_stopped,duration,offset,expected",
    [
        (0, 50, 0, True),
        (50, 10, 0, False),
        (0, 60, 10, True),
        (0, 60, 600, False),
    ],
    indirect=["time_stopped"],
)
def test_vault_lease_is_valid_accounts_for_time(duration, offset, expected):
    """
    Ensure lease validity is checked correctly and can look into the future
    """
    data = {
        "lease_id": "id",
        "renewable": False,
        "lease_duration": duration,
        "creation_time": 0,
        "expire_time": duration,
        "data": None,
    }
    res = leases.VaultLease(**data)
    assert res.is_valid_for(offset) is expected


@pytest.mark.parametrize(
    "time_stopped,duration,offset,expected",
    [
        (0, 50, 0, True),
        (50, 10, 0, False),
        (0, 60, 10, True),
        (0, 60, 600, False),
    ],
    indirect=["time_stopped"],
)
def test_vault_token_is_valid_accounts_for_time(duration, offset, expected):
    """
    Ensure token time validity is checked correctly and can look into the future
    """
    data = {
        "client_token": "id",
        "renewable": False,
        "lease_duration": duration,
        "num_uses": 0,
        "creation_time": 0,
        "expire_time": duration,
    }
    res = vault.VaultToken(**data)
    assert res.is_valid_for(offset) is expected


@pytest.mark.parametrize(
    "num_uses,uses,expected",
    [(0, 999999, True), (1, 0, True), (1, 1, False), (1, 2, False)],
)
def test_vault_token_is_valid_accounts_for_num_uses(num_uses, uses, expected):
    """
    Ensure token uses validity is checked correctly
    """
    data = {
        "client_token": "id",
        "renewable": False,
        "lease_duration": 0,
        "num_uses": num_uses,
        "creation_time": 0,
        "use_count": uses,
    }
    with patch(
        "salt.utils.vault.leases.BaseLease.is_valid_for",
        autospec=True,
        return_value=True,
    ):
        res = vault.VaultToken(**data)
        assert res.is_valid() is expected


@pytest.mark.parametrize(
    "time_stopped,duration,offset,expected",
    [
        (0, 50, 0, True),
        (50, 10, 0, False),
        (0, 60, 10, True),
        (0, 60, 600, False),
    ],
    indirect=["time_stopped"],
)
def test_vault_approle_secret_id_is_valid_accounts_for_time(duration, offset, expected):
    """
    Ensure secret ID time validity is checked correctly and can look into the future
    """
    data = {
        "secret_id": "test-secret-id",
        "renewable": False,
        "creation_time": 0,
        "expire_time": duration,
        "secret_id_num_uses": 0,
        "secret_id_ttl": duration,
    }
    res = vault.VaultSecretId(**data)
    assert res.is_valid(offset) is expected


@pytest.mark.parametrize(
    "num_uses,uses,expected",
    [(0, 999999, True), (1, 0, True), (1, 1, False), (1, 2, False)],
)
def test_vault_approle_secret_id_is_valid_accounts_for_num_uses(
    num_uses, uses, expected
):
    """
    Ensure secret ID uses validity is checked correctly
    """
    data = {
        "secret_id": "test-secret-id",
        "renewable": False,
        "creation_time": 0,
        "secret_id_ttl": 0,
        "secret_id_num_uses": num_uses,
        "use_count": uses,
    }
    with patch(
        "salt.utils.vault.leases.BaseLease.is_valid_for",
        autospec=True,
        return_value=True,
    ):
        res = vault.VaultSecretId(**data)
        assert res.is_valid() is expected


class TestLeaseStore:
    def test_get_uncached_or_invalid(self, store):
        """
        Ensure uncached or invalid leases are reported as None.
        """
        ret = store.get("test")
        assert ret is None
        store.client.post.assert_not_called()
        store.cache.flush.assert_not_called()
        store.cache.store.assert_not_called()

    def test_get_cached_valid(self, store_valid, lease):
        """
        Ensure valid leases are returned without extra behavior.
        """
        ret = store_valid.get("test")
        assert ret == lease
        store_valid.client.post.assert_not_called()
        store_valid.cache.flush.assert_not_called()
        store_valid.cache.store.assert_not_called()

    @pytest.mark.parametrize(
        "valid_for", [2000, pytest.param(2002, id="2002_renewal_leeway")]
    )
    def test_get_valid_renew_default_period(self, store_valid, lease, valid_for):
        """
        Ensure renewals are attempted by default, cache is updated accordingly
        and validity checks after renewal allow for a little leeway to account
        for latency.
        """
        ret = store_valid.get("test", valid_for=valid_for)
        lease["duration"] = lease["expire_time"] = 2000
        assert ret == lease
        store_valid.client.post.assert_called_once_with(
            "sys/leases/renew", payload={"lease_id": lease["id"]}
        )
        store_valid.cache.flush.assert_not_called()
        store_valid.cache.store.assert_called_once_with("test", ret)
        store_valid.expire_events.assert_not_called()

    def test_get_valid_renew_increment(self, store_valid, lease):
        """
        Ensure renew_increment is honored when renewing.
        """
        ret = store_valid.get("test", valid_for=1400, renew_increment=2000)
        lease["duration"] = lease["expire_time"] = 2000
        assert ret == lease
        store_valid.client.post.assert_called_once_with(
            "sys/leases/renew", payload={"lease_id": lease["id"], "increment": 2000}
        )
        store_valid.cache.flush.assert_not_called()
        store_valid.cache.store.assert_called_once_with("test", ret)
        store_valid.expire_events.assert_not_called()

    def test_get_valid_renew_increment_insufficient(self, store_valid, lease):
        """
        Ensure that when renewal_increment is set, valid_for is respected and that
        a second renewal using valid_for as increment is not attempted when the
        Vault server does not allow renewals for at least valid_for.
        If an event factory was passed, an event should be sent.
        """
        ret = store_valid.get("test", valid_for=2100, renew_increment=3000)
        assert ret is None
        store_valid.client.post.assert_has_calls(
            (
                call(
                    "sys/leases/renew",
                    payload={"lease_id": lease["id"], "increment": 3000},
                ),
                call(
                    "sys/leases/renew",
                    payload={"lease_id": lease["id"], "increment": 60},
                ),
            )
        )
        store_valid.cache.flush.assert_called_once_with("test")
        store_valid.expire_events.assert_called_once_with(
            tag="vault/lease/test/expire", data={"valid_for_less": 2100}
        )

    @pytest.mark.parametrize(
        "valid_for", [3000, pytest.param(3002, id="3002_renewal_leeway")]
    )
    def test_get_valid_renew_valid_for(
        self,
        store_valid,
        lease,
        valid_for,
        lease_renewed_response,
        lease_renewed_extended_response,
    ):
        """
        Ensure that, if renew_increment was not set and the default period
        does not yield valid_for, a second renewal is attempted by valid_for.
        There should be some leeway by default to account for latency.
        """
        store_valid.client.post.side_effect = (
            lease_renewed_response,
            lease_renewed_extended_response,
        )
        ret = store_valid.get("test", valid_for=valid_for)
        lease["duration"] = lease["expire_time"] = 3000
        assert ret == lease
        store_valid.client.post.assert_has_calls(
            (
                call("sys/leases/renew", payload={"lease_id": lease["id"]}),
                call(
                    "sys/leases/renew",
                    payload={"lease_id": lease["id"], "increment": valid_for},
                ),
            )
        )
        store_valid.cache.flush.assert_not_called()
        store_valid.cache.store.assert_called_with("test", ret)
        store_valid.expire_events.assert_not_called()

    def test_get_valid_not_renew(self, store_valid, lease):
        """
        Currently valid leases should not be returned if they undercut
        valid_for. By default, revocation should be attempted and cache
        should be flushed. If an event factory was passed, an event should be sent.
        """
        ret = store_valid.get("test", valid_for=2000, renew=False)
        assert ret is None
        store_valid.cache.store.assert_not_called()
        store_valid.client.post.assert_called_once_with(
            "sys/leases/renew", payload={"lease_id": lease["id"], "increment": 60}
        )
        store_valid.cache.flush.assert_called_once_with("test")
        store_valid.expire_events.assert_called_once_with(
            tag="vault/lease/test/expire", data={"valid_for_less": 2000}
        )

    def test_get_valid_not_flush(self, store_valid):
        """
        Currently valid leases should not be returned if they undercut
        valid_for and should not be revoked if requested so.
        If an event factory was passed, an event should be sent.
        """
        ret = store_valid.get("test", valid_for=2000, revoke=False, renew=False)
        assert ret is None
        store_valid.cache.flush.assert_not_called()
        store_valid.client.post.assert_not_called()
        store_valid.cache.store.assert_not_called()
        store_valid.expire_events.assert_called_once_with(
            tag="vault/lease/test/expire", data={"valid_for_less": 2000}
        )



import pytest

import salt.modules.test as testmod
import salt.states.group as group
import salt.utils.platform
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():

    return {
        testmod: {},
        group: {
            "__salt__": {"test.ping": testmod.ping},
            "__opts__": {"test": False},
        },
    }


def test_present_with_non_unique_gid():
    with patch(
        "salt.states.group._changes", MagicMock(side_effect=[False, {}, False])
    ), patch.dict(
        group.__salt__,
        {"group.getent": MagicMock(side_effect=[[], [{"name": "salt", "gid": 1}]])},
    ), patch.dict(
        group.__salt__, {"group.add": MagicMock(return_value=True)}
    ), patch.dict(
        group.__salt__, {"group.info": MagicMock(return_value={"things": "stuff"})}
    ):
        ret = group.present("salt", gid=1, non_unique=True)
        assert ret == {
            "changes": {"things": "stuff"},
            "comment": "New group salt created",
            "name": "salt",
            "result": True,
        }
        ret = group.present("salt", gid=1, non_unique=False)
        assert ret == {
            "changes": {},
            "comment": "Group salt is not present but gid 1 is already taken by group salt",
            "name": "salt",
            "result": False,
        }


def test_present_with_local():
    group_add_mock = MagicMock(return_value=True)
    group_info_mock = MagicMock(return_value={"things": "stuff"})
    with patch(
        "salt.states.group._changes", MagicMock(side_effect=[False, {}, False])
    ) as changes_mock, patch.dict(
        group.__salt__,
        {
            "group.getent": MagicMock(
                side_effect=[[{"name": "salt", "gid": 1}], [{"name": "salt", "gid": 1}]]
            )
        },
    ), patch.dict(
        group.__salt__, {"group.add": group_add_mock}
    ), patch.dict(
        group.__salt__, {"group.chgid": MagicMock(return_value=True)}
    ), patch.dict(
        group.__salt__, {"group.info": group_info_mock}
    ):
        ret = group.present("salt", gid=1, non_unique=True, local=True)
        assert ret["result"]
        assert changes_mock.call_args_list == [
            call("salt", 1, None, None, None, local=True),
            call("salt", 1, None, None, None, local=True),
        ]
        if salt.utils.platform.is_windows():
            group_info_mock.assert_called_once_with("salt")
            group_add_mock.assert_called_once_with(
                "salt", gid=1, system=False, non_unique=True
            )
        else:
            group_info_mock.assert_called_once_with("salt", root="/")
            group_add_mock.assert_called_once_with(
                "salt", gid=1, system=False, local=True, non_unique=True
            )


def test_present_with_existing_group_and_non_unique_gid():
    with patch(
        "salt.states.group._changes",
        MagicMock(side_effect=[{"gid": 1}, {}, {"gid": 1}, {"gid": 1}]),
    ), patch.dict(
        group.__salt__,
        {
            "group.getent": MagicMock(
                side_effect=[[{"name": "salt", "gid": 1}], [{"name": "salt", "gid": 1}]]
            )
        },
    ), patch.dict(
        group.__salt__, {"group.add": MagicMock(return_value=True)}
    ), patch.dict(
        group.__salt__, {"group.chgid": MagicMock(return_value=True)}
    ), patch.dict(
        group.__salt__, {"group.info": MagicMock(return_value={"things": "stuff"})}
    ):
        ret = group.present("salt", gid=1, non_unique=True)
        assert ret == {
            "changes": {"Final": "All changes applied successfully"},
            "comment": "The following group attributes are set to be changed:\ngid: 1\n",
            "name": "salt",
            "result": True,
        }
        ret = group.present("salt", gid=1, non_unique=False)
        assert ret == {
            "changes": {"Failed": {"gid": 1}},
            "comment": "The following group attributes are set to be changed:\n"
            "gid: 1\n"
            "Some changes could not be applied",
            "name": "salt",
            "result": False,
        }


def test_absent_with_local():
    group_delete_mock = MagicMock(return_value=True)
    group_info_mock = MagicMock(return_value={"things": "stuff"})
    with patch.dict(group.__salt__, {"group.delete": group_delete_mock}), patch.dict(
        group.__salt__, {"group.info": group_info_mock}
    ):
        ret = group.absent("salt", local=True)
        assert ret == {
            "changes": {"salt": ""},
            "comment": "Removed group salt",
            "name": "salt",
            "result": True,
        }
        if salt.utils.platform.is_windows():
            group_info_mock.assert_called_once_with("salt")
            group_delete_mock.assert_called_once_with("salt")
        else:
            group_info_mock.assert_called_once_with("salt", root="/")
            group_delete_mock.assert_called_once_with("salt", local=True)



"""
integration tests for mac_system
"""

import logging

import pytest
from saltfactories.utils import random_string

from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_unless_on_darwin,
    pytest.mark.skip_if_binaries_missing("systemsetup"),
]


@pytest.fixture
def service(modules):
    return modules.service


@pytest.fixture
def system(modules):
    return modules.system


@pytest.fixture
def _remote_login_cleanup(system, grains):
    if grains["osmajorrelease"] >= 13:
        pytest.skip("SKipping until we figure out how to have full dist access")

    remote_login_enabled = system.get_remote_login()
    try:
        yield
    finally:
        if system.get_remote_login() != remote_login_enabled:
            system.set_remote_login(remote_login_enabled)


@pytest.fixture
def _remote_events_cleanup(system, grains):
    if grains["osmajorrelease"] >= 13:
        pytest.skip("SKipping until we figure out how to have full dist access")

    remote_events_enabled = system.get_remote_events()
    try:
        yield
    finally:
        if system.get_remote_events() != remote_events_enabled:
            system.set_remote_events(remote_events_enabled)


@pytest.fixture
def subnet_name(system):
    subnet_name = system.get_subnet_name()
    try:
        yield random_string("subnet-", lowercase=False)
    finally:
        if system.get_subnet_name() != subnet_name:
            system.set_subnet_name(subnet_name)


@pytest.fixture
def _keyboard_cleanup(system):
    keyboard_disabled = system.get_disable_keyboard_on_lock()
    try:
        yield
    finally:
        if system.get_disable_keyboard_on_lock() != keyboard_disabled:
            system.set_disable_keyboard_on_lock(keyboard_disabled)


@pytest.fixture
def computer_name(system):
    computer_name = system.get_computer_name()
    try:
        yield random_string("cmptr-", lowercase=False)
    finally:
        if system.get_computer_name() != computer_name:
            system.set_computer_name(computer_name)


@pytest.mark.usefixtures("_remote_login_cleanup")
def test_get_set_remote_login(system):
    """
    Test system.get_remote_login
    Test system.set_remote_login
    """
    # Normal Functionality
    ret = system.set_remote_login(True)
    assert ret

    ret = system.get_remote_login()
    assert ret

    ret = system.set_remote_login(False)
    assert ret

    ret = system.get_remote_login()
    assert not ret

    # Test valid input
    ret = system.set_remote_login(True)
    assert ret

    ret = system.set_remote_login(False)
    assert ret

    ret = system.set_remote_login("yes")
    assert ret

    ret = system.set_remote_login("no")
    assert ret

    ret = system.set_remote_login("On")
    assert ret

    ret = system.set_remote_login("Off")
    assert ret

    ret = system.set_remote_login(1)
    assert ret

    ret = system.set_remote_login(0)
    assert ret

    # Test invalid input
    with pytest.raises(SaltInvocationError) as exc:
        system.set_remote_login("spongebob")
        assert "Invalid String Value for Enabled" in str(exc.value)


@pytest.mark.skip_initial_gh_actions_failure
@pytest.mark.usefixtures("_remote_events_cleanup")
def test_get_set_remote_events(system):
    """
    Test system.get_remote_events
    Test system.set_remote_events
    """
    # Normal Functionality
    ret = system.set_remote_events(True)
    assert ret

    ret = system.get_remote_events()
    assert ret

    ret = system.set_remote_events(False)
    assert ret

    ret = not system.get_remote_events()
    assert not ret

    # Test valid input
    ret = system.set_remote_events(True)
    assert ret

    ret = system.set_remote_events(False)
    assert ret

    ret = system.set_remote_events("yes")
    assert ret

    ret = system.set_remote_events("no")
    assert ret

    ret = system.set_remote_events("On")
    assert ret

    ret = system.set_remote_events("Off")
    assert ret

    ret = system.set_remote_events(1)
    assert ret

    ret = system.set_remote_events(0)
    assert ret

    # Test invalid input
    with pytest.raises(CommandExecutionError) as exc:
        system.set_remote_events("spongebob")
        assert "Invalid String Value for Enabled" in str(exc.value)


def test_get_set_subnet_name(system, subnet_name):
    """
    Test system.get_subnet_name
    Test system.set_subnet_name
    """
    ret = system.set_subnet_name(subnet_name)
    assert ret

    ret = system.get_subnet_name()
    assert ret == subnet_name


@pytest.mark.skip_initial_gh_actions_failure
def test_get_list_startup_disk(system):
    """
    Test system.get_startup_disk
    Test system.list_startup_disks
    Don't know how to test system.set_startup_disk as there's usually only
    one startup disk available on a system
    """
    # Test list and get
    ret = system.list_startup_disks()
    assert isinstance(ret, list)

    startup_disk = system.get_startup_disk()
    assert startup_disk in ret

    # Test passing set a bad disk
    with pytest.raises(SaltInvocationError) as exc:
        system.set_startup_disk("spongebob")
        assert "Invalid value passed for path." in str(exc.value)


@pytest.mark.skip(reason="Skip this test until mac fixes it.")
def test_get_set_restart_delay(system):
    """
    Test system.get_restart_delay
    Test system.set_restart_delay
    system.set_restart_delay does not work due to an apple bug, see docs
    may need to disable this test as we can't control the delay value
    """
    # Normal Functionality
    ret = system.set_restart_delay(90)
    assert ret

    ret = system.get_restart_delay()
    assert ret == "90 seconds"

    # Pass set bad value for seconds
    with pytest.raises(CommandExecutionError) as exc:
        system.set_restart_delay(70)
        assert "Invalid value passed for seconds." in str(exc.value)


@pytest.mark.usefixtures("_keyboard_cleanup")
def test_get_set_disable_keyboard_on_lock(system):
    """
    Test system.get_disable_keyboard_on_lock
    Test system.set_disable_keyboard_on_lock
    """
    # Normal Functionality
    ret = system.set_disable_keyboard_on_lock(True)
    assert ret

    ret = system.get_disable_keyboard_on_lock()
    assert ret

    ret = system.set_disable_keyboard_on_lock(False)
    assert ret

    ret = system.get_disable_keyboard_on_lock()
    assert not ret

    # Test valid input
    ret = system.set_disable_keyboard_on_lock(True)
    assert ret

    ret = system.set_disable_keyboard_on_lock(False)
    assert ret

    ret = system.set_disable_keyboard_on_lock("yes")
    assert ret

    ret = system.set_disable_keyboard_on_lock("no")
    assert ret

    ret = system.set_disable_keyboard_on_lock("On")
    assert ret

    ret = system.set_disable_keyboard_on_lock("Off")
    assert ret

    ret = system.set_disable_keyboard_on_lock(1)
    assert ret

    ret = system.set_disable_keyboard_on_lock(0)
    assert ret

    # Test invalid input
    with pytest.raises(SaltInvocationError) as exc:
        system.set_disable_keyboard_on_lock("spongebob")
        assert "Invalid String Value for Enabled" in str(exc.value)


@pytest.mark.skip(reason="Skip this test until mac fixes it.")
def test_get_set_boot_arch(system):
    """
    Test system.get_boot_arch
    Test system.set_boot_arch
    system.set_boot_arch does not work due to an apple bug, see docs
    may need to disable this test as we can't set the boot architecture
    """
    # Normal Functionality
    ret = system.set_boot_arch("i386")
    assert ret

    ret = system.get_boot_arch()
    assert ret == "i386"

    ret = system.set_boot_arch("default")
    assert ret

    ret = system.get_boot_arch()
    assert ret == "default"

    # Test invalid input
    with pytest.raises(CommandExecutionError) as exc:
        system.set_boot_arch("spongebob")
        assert "Invalid value passed for arch" in str(exc.value)


# A similar test used to be skipped on py3 due to 'hanging', if we see
# something similar again we may want to skip this gain until we
# investigate
# @pytest.mark.skipif(salt.utils.platform.is_darwin() and six.PY3, reason='This test hangs on OS X on Py3.  Skipping until #53566 is merged.')
@pytest.mark.destructive_test
def test_get_set_computer_name(system, computer_name):
    """
    Test system.get_computer_name
    Test system.set_computer_name
    """
    current_computer_name = system.get_computer_name()
    assert current_computer_name
    assert current_computer_name != computer_name

    ret = system.set_computer_name(computer_name)
    assert ret

    ret = system.get_computer_name()
    assert ret == computer_name
